import os

import numpy as np

import dtcwt
from dtcwt.numpy import Transform2d
from dtcwt.registration import *

import tests.datasets as datasets

def setup():
    global f1, f2
    f1, f2 = datasets.regframes('traffic')

def test_frames_loaded():
    assert f1.shape == (576, 768)
    assert f1.min() >= 0
    assert f1.max() <= 1
    assert f1.dtype == np.float64

    assert f2.shape == (576, 768)
    assert f2.min() >= 0
    assert f2.max() <= 1
    assert f2.dtype == np.float64

def test_estimatereg():
    nlevels = 6
    trans = Transform2d()
    t1 = trans.forward(f1, nlevels=nlevels)
    t2 = trans.forward(f2, nlevels=nlevels)
    avecs = estimatereg(t1, t2)

    # Make sure warped frame 1 has lower mean overlap error than non-warped
    warped_f1 = warp(f1, avecs, method='bilinear')
    assert np.mean(np.abs(warped_f1 - f2)) < np.mean(np.abs(f1-f2))




"""Functions to load standard wavelet coefficients.

"""
from __future__ import absolute_import

import os
from numpy import load

DATADIR = os.path.join(os.path.dirname(__file__), 'data')

COEFF_CACHE = {}

def _load_from_file(basename, varnames):
    filename = os.path.join(DATADIR, basename + '.npz')

    try:
        mat = COEFF_CACHE[filename]
    except KeyError:
        mat = load(filename)
        COEFF_CACHE[filename] = mat

    try:
        return tuple(mat[k] for k in varnames)
    except KeyError:
        raise ValueError('Wavelet does not define ({0}) coefficients'.format(', '.join(varnames)))

def biort(name):
    """Load level 1 wavelet by name.

    :param name: a string specifying the wavelet family name
    :returns: a tuple of vectors giving filter coefficients

    =============  ============================================
    Name           Wavelet
    =============  ============================================
    antonini       Antonini 9,7 tap filters.
    legall         LeGall 5,3 tap filters.
    near_sym_a     Near-Symmetric 5,7 tap filters.
    near_sym_b     Near-Symmetric 13,19 tap filters.
    near_sym_b_bp  Near-Symmetric 13,19 tap filters + BP filter
    =============  ============================================

    Return a tuple whose elements are a vector specifying the h0o, g0o, h1o and
    g1o coefficients.

    See :ref:`rot-symm-wavelets` for an explanation of the ``near_sym_b_bp``
    wavelet filters.

    :raises IOError: if name does not correspond to a set of wavelets known to the library.
    :raises ValueError: if name specifies a :py:func:`dtcwt.coeffs.qshift` wavelet.

    """
    if name == 'near_sym_b_bp':
        return _load_from_file(name, ('h0o', 'g0o', 'h1o', 'g1o', 'h2o', 'g2o'))
    else:
        return _load_from_file(name, ('h0o', 'g0o', 'h1o', 'g1o'))

def qshift(name):
    """Load level >=2 wavelet by name,

    :param name: a string specifying the wavelet family name
    :returns: a tuple of vectors giving filter coefficients

    ============ ============================================
    Name         Wavelet
    ============ ============================================
    qshift_06    Quarter Sample Shift Orthogonal (Q-Shift) 10,10 tap filters,
                 (only 6,6 non-zero taps).
    qshift_a     Q-shift 10,10 tap filters,
                 (with 10,10 non-zero taps, unlike qshift_06).
    qshift_b     Q-Shift 14,14 tap filters.
    qshift_c     Q-Shift 16,16 tap filters.
    qshift_d     Q-Shift 18,18 tap filters.
    qshift_b_bp  Q-Shift 18,18 tap filters + BP
    ============ ============================================

    Return a tuple whose elements are a vector specifying the h0a, h0b, g0a,
    g0b, h1a, h1b, g1a and g1b coefficients.

    See :ref:`rot-symm-wavelets` for an explanation of the ``qshift_b_bp``
    wavelet filters.

    :raises IOError: if name does not correspond to a set of wavelets known to the library.
    :raises ValueError: if name specifies a :py:func:`dtcwt.coeffs.biort` wavelet.

    """
    if name == 'qshift_b_bp':
        return _load_from_file(name, ('h0a', 'h0b', 'g0a', 'g0b', 'h1a', 'h1b', 'g1a', 'g1b', 'h2a', 'h2b', 'g2a','g2b'))
    else:
        return _load_from_file(name, ('h0a', 'h0b', 'g0a', 'g0b', 'h1a', 'h1b', 'g1a', 'g1b'))

# vim:sw=4:sts=4:et



import os
import numpy as np

def regframes(name):
    """Load the *name* registration dataset and return source and reference frame."""
    frames = np.load(os.path.join(os.path.dirname(__file__), name + '.npz'))
    return frames['f1'], frames['f2']

def mandrill():
    """Return the "mandrill" test image."""
    return np.load(os.path.join(os.path.dirname(__file__), 'mandrill.npz'))['mandrill']



from __future__ import absolute_import

try:
    import tensorflow as tf
except ImportError:
    # The lack of tensorflow will be caught by the low-level routines.
    pass


class Pyramid(object):
    """A tensorflow representation of a transform domain signal.

    An interface-compatible version of
    :py:class:`dtcwt.Pyramid` where the initialiser
    arguments are assumed to be :py:class:`tf.Variable` instances.

    The attributes defined in :py:class:`dtcwt.Pyramid`
    are implemented via properties. The original tf arrays may be accessed
    via the ``..._op(s)`` attributes.

    .. py:attribute:: lowpass_op

        A tensorflow tensor that can be evaluated in a session to return
        the coarsest scale lowpass signal for the input, X.

    .. py:attribute:: highpasses_op

        A tuple of tensorflow tensors, where each element is the complex
        subband coefficients for corresponding scales finest to coarsest.

    .. py:attribute:: scales_ops

        *(optional)* A tuple where each element is a tensorflow tensor
        containing the lowpass signal for corresponding scales finest to
        coarsest. This is not required for the inverse and may be *None*.
    """
    def __init__(self, lowpass, highpasses, scales=None, numpy=False):
        self.lowpass_op = lowpass
        self.highpasses_ops = highpasses
        self.scales_ops = scales
        self.numpy = numpy

    @property
    def lowpass(self):
        if not hasattr(self, '_lowpass'):
            if self.lowpass_op is None:
                self._lowpass = None
            else:
                with tf.Session() as sess:
                    sess.run(tf.global_variables_initializer())
                    self._lowpass = sess.run(self.lowpass_op)
        return self._lowpass

    @property
    def highpasses(self):
        if not hasattr(self, '_highpasses'):
            if self.highpasses_ops is None:
                self._highpasses = None
            else:
                with tf.Session() as sess:
                    sess.run(tf.global_variables_initializer())
                    self._highpasses = \
                        tuple(sess.run(x) for x in self.highpasses_ops)
        return self._highpasses

    @property
    def scales(self):
        if not hasattr(self, '_scales'):
            if self.scales_ops is None:
                self._scales = None
            else:
                with tf.Session() as sess:
                    sess.run(tf.global_variables_initializer())
                    self._scales = tuple(sess.run(x) for x in self.scales_ops)
        return self._scales



#!/usr/bin/env python

"""
Register neighbouring frames of video and save inter-frame transform
parameters to a file.

Usage:
    register_images.py [options] <prevframe> <nextframe> <output>
    register_images.py (-h | --help)

Options:

    --cl        Attempt to use OpenCL where possible.

"""

import logging

from docopt import docopt
import dtcwt
from dtcwt.opencl import Transform2d as CLTransform2d
from dtcwt.numpy import Transform2d as NumPyTransform2d
import dtcwt.registration as reg
import dtcwt.sampling
from PIL import Image # Use 'Pillow', the PIL fork
import numpy as np
import tables

# Parse command line options
OPTS = docopt(__doc__)

# Set logging options
logging.basicConfig(level=logging.INFO)

def avecs_for_pair(prev, next_):
    trans = CLTransform2d() if OPTS['--cl'] else NumPyTransform2d()
    t1 = trans.forward(prev, nlevels=5)
    t2 = trans.forward(next_, nlevels=5)
    return reg.estimatereg(t1, t2)

class Metadata(tables.IsDescription):
    previmpath = tables.StringCol(512)
    nextimpath = tables.StringCol(512)

def main():
    logging.info('Launched')

    logging.info('Loading "prev" image from "{0}"'.format(OPTS['<prevframe>']))
    pim = np.array(Image.open(OPTS['<prevframe>']).convert('L')) / 255.0

    logging.info('Loading "next" image from "{0}"'.format(OPTS['<nextframe>']))
    nim = np.array(Image.open(OPTS['<nextframe>']).convert('L')) / 255.0

    logging.info('Estimating registration')
    avecs = avecs_for_pair(pim, nim)

    logging.info('Calculating velocity field')
    vxs, vys = dtcwt.registration.velocityfield(avecs, avecs.shape[:2], method='bilinear')

    logging.info('Saving result to {0}'.format(OPTS['<output>']))
    np.savez_compressed(OPTS['<output>'], avecs=avecs, vxs=vxs, vys=vys)

if __name__ == '__main__':
    main()



import numpy as np
from dtcwt.coeffs import biort, qshift
from dtcwt.numpy.lowlevel import colfilter as np_colfilter
from importlib import import_module

from tests.util import skip_if_no_tf
import tests.datasets as datasets


@skip_if_no_tf
def test_setup():
    global mandrill, mandrill_t, tf, colfilter
    tf = import_module('tensorflow')
    lowlevel = import_module('dtcwt.tf.lowlevel')
    colfilter = getattr(lowlevel, 'colfilter')

    mandrill = datasets.mandrill()
    mandrill_t = tf.expand_dims(tf.constant(mandrill, dtype=tf.float32),axis=0)


@skip_if_no_tf
def test_mandrill_loaded():
    assert mandrill.shape == (512, 512)
    assert mandrill.min() >= 0
    assert mandrill.max() <= 1
    assert mandrill.dtype == np.float32
    assert mandrill_t.get_shape() == (1, 512, 512)


@skip_if_no_tf
def test_odd_size():
    y_op = colfilter(mandrill_t, [-1,2,-1])
    assert y_op.get_shape()[1:] == mandrill.shape


@skip_if_no_tf
def test_even_size():
    y_op = colfilter(mandrill_t, [-1,-1])
    assert y_op.get_shape()[1:] == (mandrill.shape[0]+1, mandrill.shape[1])


@skip_if_no_tf
def test_qshift():
    h = qshift('qshift_a')[0]
    y_op = colfilter(mandrill_t, h)
    assert y_op.get_shape()[1:] == (mandrill.shape[0]+1, mandrill.shape[1])


@skip_if_no_tf
def test_biort():
    h = biort('antonini')[0]
    y_op = colfilter(mandrill_t, h)
    assert y_op.get_shape()[1:] == mandrill.shape


@skip_if_no_tf
def test_even_size_batch():
    zero_t = tf.zeros([1, mandrill.shape[0], mandrill.shape[1]], tf.float32)
    y_op = colfilter(zero_t, [-1,1])
    assert y_op.get_shape()[1:] == (mandrill.shape[0]+1, mandrill.shape[1])
    with tf.Session() as sess:
        y = sess.run(y_op)
    assert not np.any(y[:] != 0.0)


@skip_if_no_tf
def test_equal_small_in():
    h = qshift('qshift_b')[0]
    im = mandrill[0:4,0:4]
    im_t = tf.expand_dims(tf.constant(im, tf.float32), axis=0)
    ref = np_colfilter(im, h)
    y_op = colfilter(im_t, h)
    with tf.Session() as sess:
        y = sess.run(y_op)
    np.testing.assert_array_almost_equal(y[0], ref, decimal=4)


@skip_if_no_tf
def test_equal_numpy_biort1():
    h = biort('near_sym_b')[0]
    ref = np_colfilter(mandrill, h)
    y_op = colfilter(mandrill_t, h)
    with tf.Session() as sess:
        y = sess.run(y_op)
    np.testing.assert_array_almost_equal(y[0], ref, decimal=4)


@skip_if_no_tf
def test_equal_numpy_biort2():
    h = biort('near_sym_b')[0]
    im = mandrill[52:407, 30:401]
    im_t = tf.expand_dims(tf.constant(im, tf.float32), axis=0)
    ref = np_colfilter(im, h)
    y_op = colfilter(im_t, h)
    with tf.Session() as sess:
        y = sess.run(y_op)
    np.testing.assert_array_almost_equal(y[0], ref, decimal=4)


@skip_if_no_tf
def test_equal_numpy_qshift1():
    h = qshift('qshift_c')[0]
    ref = np_colfilter(mandrill, h)
    y_op = colfilter(mandrill_t, h)
    with tf.Session() as sess:
        y = sess.run(y_op)
    np.testing.assert_array_almost_equal(y[0], ref, decimal=4)


@skip_if_no_tf
def test_equal_numpy_qshift2():
    h = qshift('qshift_c')[0]
    im = mandrill[52:407, 30:401]
    im_t = tf.expand_dims(tf.constant(im, tf.float32), axis=0)
    ref = np_colfilter(im, h)
    y_op = colfilter(im_t, h)
    with tf.Session() as sess:
        y = sess.run(y_op)
    np.testing.assert_array_almost_equal(y[0], ref, decimal=4)

# vim:sw=4:sts=4:et



import os
import pytest

from importlib import import_module

from .util import skip_if_no_tf
from dtcwt.utils import unpack
import dtcwt
import dtcwt.compat

PRECISION_DECIMAL = 5


@skip_if_no_tf
def setup():
    global tf
    tf = import_module('tensorflow')
    dtcwt.push_backend('tf')

    # Make sure we run tests on cpu rather than gpus
    os.environ["CUDA_VISIBLE_DEVICES"] = ""


@skip_if_no_tf
@pytest.mark.parametrize("nlevels, include_scale", [
    (2,False),
    (2,True),
    (4,False),
    (3,True)
])
def test_scales(nlevels, include_scale):
    in_ = tf.placeholder(tf.float32, [512, 512])
    t = dtcwt.Transform2d()

    p = t.forward(in_, nlevels, include_scale)

    # At level 1, the lowpass output will be the same size as the input. At
    # levels above that, it will be half the size per level
    extent = 512 * 2**(-(nlevels-1))
    assert p.lowpass_op.get_shape().as_list() == [extent, extent]
    assert p.lowpass_op.dtype == tf.float32

    for i in range(nlevels):
        extent = 512 * 2**(-(i+1))
        assert (p.highpasses_ops[i].get_shape().as_list() ==
                [extent, extent, 6])
        assert (p.highpasses_ops[i].dtype ==
                tf.complex64)
        if include_scale:
            assert (p.scales_ops[i].get_shape().as_list() ==
                    [2*extent, 2*extent])
            assert p.scales_ops[i].dtype == tf.float32


@skip_if_no_tf
@pytest.mark.parametrize("nlevels, include_scale", [
    (2,False),
    (2,True),
    (4,False),
    (3,True)
])
def test_2d_input_tuple(nlevels, include_scale):
    in_ = tf.placeholder(tf.float32, [512, 512])
    t = dtcwt.Transform2d()
    if include_scale:
        Yl, Yh, Yscale = unpack(t.forward(in_, nlevels, include_scale), 'tf')
    else:
        Yl, Yh = unpack(t.forward(in_, nlevels, include_scale), 'tf')

    # At level 1, the lowpass output will be the same size as the input. At
    # levels above that, it will be half the size per level
    extent = 512 * 2**(-(nlevels-1))
    assert Yl.get_shape().as_list() == [extent, extent]
    assert Yl.dtype == tf.float32

    for i in range(nlevels):
        extent = 512 * 2**(-(i+1))
        assert Yh[i].get_shape().as_list() == [extent, extent, 6]
        assert Yh[i].dtype == tf.complex64
        if include_scale:
            assert Yscale[i].get_shape().as_list() == [2*extent, 2*extent]
            assert Yscale[i].dtype == tf.float32


@skip_if_no_tf
@pytest.mark.parametrize("nlevels, include_scale, batch_size", [
    (2,False,None),
    (2,True,10),
    (4,False,None),
    (3,True,2)
])
def test_batch_input(nlevels, include_scale, batch_size):
    in_ = tf.placeholder(tf.float32, [batch_size, 512, 512])
    t = dtcwt.Transform2d()
    p = t.forward_channels(in_, "nhw", nlevels, include_scale)

    # At level 1, the lowpass output will be the same size as the input. At
    # levels above that, it will be half the size per level
    extent = 512 * 2**(-(nlevels-1))
    assert p.lowpass_op.get_shape().as_list() == [batch_size, extent, extent]
    assert p.lowpass_op.dtype == tf.float32

    for i in range(nlevels):
        extent = 512 * 2**(-(i+1))
        assert (p.highpasses_ops[i].get_shape().as_list() ==
                [batch_size, extent, extent, 6])
        assert p.highpasses_ops[i].dtype == tf.complex64
        if include_scale:
            assert (p.scales_ops[i].get_shape().as_list() ==
                    [batch_size, 2*extent, 2*extent])
            assert p.scales_ops[i].dtype == tf.float32


@skip_if_no_tf
@pytest.mark.parametrize("nlevels, include_scale, batch_size", [
    (2,False,None),
    (2,True,10),
    (4,False,None),
    (3,True,2)
])
def test_batch_input_tuple(nlevels, include_scale, batch_size):
    in_ = tf.placeholder(tf.float32, [batch_size, 512, 512])
    t = dtcwt.Transform2d()

    if include_scale:
        Yl, Yh, Yscale = unpack(
            t.forward_channels(in_, "nhw", nlevels, include_scale), "tf")
    else:
        Yl, Yh = unpack(
            t.forward_channels(in_, "nhw", nlevels, include_scale), "tf")

    # At level 1, the lowpass output will be the same size as the input. At
    # levels above that, it will be half the size per level
    extent = 512 * 2**(-(nlevels-1))
    assert Yl.get_shape().as_list() == [batch_size, extent, extent]
    assert Yl.dtype == tf.float32

    for i in range(nlevels):
        extent = 512 * 2**(-(i+1))
        assert Yh[i].get_shape().as_list() == [batch_size, extent, extent, 6]
        assert Yh[i].dtype == tf.complex64
        if include_scale:
            assert (Yscale[i].get_shape().as_list() ==
                    [batch_size, 2*extent, 2*extent])
            assert Yscale[i].dtype == tf.float32


@skip_if_no_tf
@pytest.mark.parametrize("nlevels, channels", [
    (2,5),
    (2,2),
    (4,10),
    (3,6)
])
def test_multichannel(nlevels, channels):
    in_ = tf.placeholder(tf.float32, [None, 512, 512, channels])
    t = dtcwt.Transform2d()
    Yl, Yh, Yscale = unpack(
        t.forward_channels(in_, "nhwc", nlevels, include_scale=True), "tf")
    # At level 1, the lowpass output will be the same size as the input. At
    # levels above that, it will be half the size per level
    extent = 512 * 2**(-(nlevels-1))
    assert Yl.get_shape().as_list() == [None, extent, extent, channels]
    assert Yl.dtype == tf.float32

    for i in range(nlevels):
        extent = 512 * 2**(-(i+1))
        assert (Yh[i].get_shape().as_list() ==
                [None, extent, extent, channels, 6])
        assert Yh[i].dtype == tf.complex64
        assert Yscale[i].get_shape().as_list() == [
            None, 2*extent, 2*extent, channels]
        assert Yscale[i].dtype == tf.float32



from __future__ import absolute_import

import numpy as np

from dtcwt.sampling import upsample_highpass, upsample

__all__ = [ 'find_keypoints' ]

def find_keypoints(highpass_highpasses, method=None,
        alpha=1.0, beta=0.4, kappa=1.0/6.0,
        threshold=None, max_points=None,
        upsample_keypoint_energy=None, upsample_highpasses=None,
        refine_positions=True, skip_levels=1):
    """
    :param highpass_highpasses: (NxMx6) matrix of highpass subband images
    :param method: *(optional)* string specifying which keypoint energy method to use
    :param alpha: *(optional)* scale parameter for ``'fauqueur'`` method
    :param beta: *(optional)* shape parameter for ``'fauqueur'`` method
    :param kappa: *(optiona)* suppression parameter for ``'kingsbury'`` method
    :param threshold: *(optional)* minimum keypoint energy of returned keypoints
    :param max_points: *(optional)* maximum number of keypoints to return
    :param upsample_keypoint_energy: is non-None, a string specifying a method used to upscale the keypoint energy map before finding keypoints
    :param upsample_subands: is non-None, a string specifying a method used to upscale the subband image before finding keypoints
    :param refine_positions: *(optional)* should the keypoint positions be refined to sub-pixel accuracy
    :param skip_levels: *(optional)* number of levels of the transform to ignore before looking for keypoints

    :returns: (Px4) array of P keypoints in image co-ordinates

    .. warning::

        The interface and behaviour of this function is the subject of an open
        research project. It is provided in this release as a preview of
        forthcoming functionality but it is subject to change between releases.

    The rows of the returned keypoint array give the x co-ordinate, y
    co-ordinate, scale and keypoint energy. The rows are sorted in order of
    decreasing keypoint energy.

    If *refine_positions* is ``True`` then the positions (and energy) of the
    keypoints will be refined to sub-pixel accuracy by fitting a quadratic
    patch. If *refine_positions* is ``False`` then the keypoint locations will
    be those corresponding directly to pixel-wise maxima of the subband images.

    The *max_points* and *threshold* parameters are cumulative: if both are
    specified then the *max_points* greatest energy keypoints with energy
    greater than *threshold* will be returned.

    Usually the keypoint energies returned from the finest scale level are
    dominated by noise and so one usually wants to specify *skip_levels* to be
    1 or 2. If *skip_levels* is 0 then all levels will be used to compute
    keypoint energy.

    The *upsample_highpasses* and *upsample_keypoint_energy* parameters are used
    to control whether the individual subband coefficients and/org the keypoint
    energy map are upscaled by 2 before finding keypoints. If these parameters
    are None then no corresponding upscaling is performed. If non-None they
    specify the upscale method as outlined in
    :py:func:`dtcwt.sampling.upsample`.

    If *method* is ``None``, the default ``'fauqueur'`` method is used.

    =========== ======================================= ======================
    Name        Description                             Parameters used
    =========== ======================================= ======================
    fauqueur    Geometric mean of absolute values[1]    *alpha*, *beta*
    bendale     Minimum absolute value[2]               none
    kingsbury   Cross-product of orthogonal highpasses    *kappa*
    =========== ======================================= ======================

    [1] Julien Fauqueur, Nick Kingsbury, and Ryan Anderson. *Multiscale
    Keypoint Detection using the Dual-Tree Complex Wavelet Transform*. 2006
    International Conference on Image Processing, pages 1625-1628, October
    2006. ISSN 1522-4880. doi: 10.1109/ICIP.2006.312656.
    http://ieeexplore.ieee.org/lpdocs/epic03/wrapper.htm?arnumber=4106857.

    [2] Pashmina Bendale, Bill Triggs, and Nick Kingsbury. *Multiscale Keypoint
    Analysis based on Complex Wavelets*. In British Machine Vision Con-ference
    (BMVC), 2010.
    http://www-sigproc.eng.cam.ac.uk/~pb397/publications/BTK_BMVC_2010_abstract.pdf.
    """
    # Set default method
    if method is None:
        method = 'fauqueur'

    # Skip levels
    highpass_highpasses = highpass_highpasses[skip_levels:]

    # Compute contribution to scale from upsampling
    upsample_scale = 1
    if upsample_highpasses is not None:
        upsample_scale <<= 1
    if upsample_keypoint_energy is not None:
        upsample_scale <<= 1

    # Find keypoint energy map for each level
    kp_energies = []
    for scale, subband in enumerate(highpass_highpasses):
        if upsample_highpasses is not None:
            subband = upsample_highpass(subband, upsample_highpasses)

        if method == 'fauqueur':
            kp_energies.append(_keypoint_energy_fauqueur(subband, alpha, beta, scale))
        elif method == 'bendale':
            kp_energies.append(_keypoint_energy_bendale(subband))
        elif method == 'kingsbury':
            kp_energies.append(_keypoint_energy_kingsbury(subband, kappa))
        elif method == 'gale':
            kp_energies.append(_keypoint_energy_gale(subband))
        else:
            raise ValueError('Unknown method: {0}'.format(method))

        if upsample_keypoint_energy is not None:
            kp_energies[-1] = upsample(kp_energies[-1], upsample_keypoint_energy)

    # Find keypoints for each level
    kps = None
    for level_idx, kp_energy in enumerate(kp_energies):
        kp_scale = 2**(level_idx+1+skip_levels) / float(upsample_scale)
        kp_rows, kp_cols, kp_energies = _kp_energy_maxima(kp_energy, threshold=threshold, refine=refine_positions)

        # Scaling is a bit non-trivial. If the subband has pixel coords in range {0, .., M-1} then it has extent
        # (-0.5, M-0.5]. If we need to scale the pixel size by kp_scale then the final image will have extent
        # (-0.5, kp_scale*M-0.5]. So we need a linear function which maps -0.5 -> -0.5 and M-0.5 -> kp_scale*M-0.5
        # such a function is x -> kp_scale * (x+0.5) - 0.5

        level_kps = np.array((
            (kp_cols+0.5)*kp_scale-0.5, (kp_rows+0.5)*kp_scale-0.5,
            kp_scale*np.ones(kp_cols.shape[0]), kp_energies)).T

        if kps is None:
            kps = level_kps
        else:
            kps = np.vstack((kps, level_kps))

    # Sort keypoints
    sorted_indices = np.argsort(kps[:, 3])
    kps = kps[sorted_indices[::-1],:]

    # Truncate if necessary
    if max_points is not None:
        kps = kps[:max_points]

    # Return keypoints
    return kps

def _keypoint_energy_fauqueur(subband, alpha, beta, scale):
    return (alpha**(scale+1)) * np.power(np.maximum(0, np.product(np.abs(subband), axis=2)), beta)

def _keypoint_energy_bendale(subband):
    return np.min(np.abs(subband), axis=2)

def _keypoint_energy_kingsbury(subband, kappa=1.0/6.0, epsilon=1e-8):
    abs_Y = np.abs(subband)
    A = np.sqrt(np.sum(abs_Y*abs_Y, axis=2))
    B = np.sum(abs_Y[:,:,:3] * abs_Y[:,:,3:], axis=2)

    # The max(0, ...) is not part of the original energy calculation but we use
    # it to avoid finding false maxima in no-threshold cases.
    return np.maximum(0, B/np.maximum(epsilon, A) - kappa*A)

def _keypoint_energy_gale(subband):
    raise NotImplementedError('not implemented yet')

def _nullspace(A, atol=1e-13, rtol=0):
    """Compute an approximate basis for the nullspace of A.

    The algorithm used by this function is based on the singular value
    decomposition of `A`.

    Parameters
    ----------
    A : ndarray
        A should be at most 2-D.  A 1-D array with length k will be treated
        as a 2-D with shape (1, k)
    atol : float
        The absolute tolerance for a zero singular value.  Singular values
        smaller than `atol` are considered to be zero.
    rtol : float
        The relative tolerance.  Singular values less than rtol*smax are
        considered to be zero, where smax is the largest singular value.

    If both `atol` and `rtol` are positive, the combined tolerance is the
    maximum of the two; that is::
        tol = max(atol, rtol * smax)
    Singular values smaller than `tol` are considered to be zero.

    Return value
    ------------
    ns : ndarray
        If `A` is an array with shape (m, k), then `ns` will be an array
        with shape (k, n), where n is the estimated dimension of the
        nullspace of `A`.  The columns of `ns` are a basis for the
        nullspace; each element in numpy.dot(A, ns) will be approximately
        zero.
    """

    A = np.atleast_2d(A)
    u, s, vh = np.linalg.svd(A)
    tol = max(atol, rtol * s[0])
    nnz = (s >= tol).sum()
    ns = vh[nnz:].conj().T
    return ns

def _kp_energy_maxima(X, threshold=None, refine=True):
    # If no threshold is provided, choose one which all keypoints will pass
    if threshold is None:
        threshold = X.min() - 1

    # Compute local maximum image
    maxima = np.ones_like(X) * threshold
    for dx in (-1,0,1):
        for dy in (-1,0,1):
            maxima[1:-2,1:-2] = np.maximum(maxima[1:-2,1:-2], X[1+dy:X.shape[0]-2+dy, 1+dx:X.shape[1]-2+dx])

    # This will be used to store the values of local maxima
    lm_values = []

    # This will be used to store the refined positions
    lm_refined_rows, lm_refined_cols = [], []

    # Find local maxima
    lm_rows, lm_cols = np.nonzero(maxima == X)

    if refine:
        # Taylor series of I(x) around x_0 is I(x) ~= I(x_o) + dI/dx x + dI^2/dx^2 x^2 + ...
        # maximum is at differential is zero or:
        # 0 = dI/dx + 2 dI^2/dx^2 x => x = -dI/dx * (2*dI^2/dx^2)^-1

        # Form the various gradient images for X
        dXdy, dXdx = np.gradient(X)
        dX2dxdy, dX2dx2 = np.gradient(dXdx)
        dX2dy2, dX2dydx = np.gradient(dXdy)

        a_im = np.dstack((
           dX2dx2, dX2dy2, dX2dxdy, dXdx, dXdy, X,
        ))

    # Calculate a vectors for each neighbourhood
    for r, c in zip(lm_rows, lm_cols):
        if refine:
            a = a_im[r,c,:]
            A = np.array(((2*a[0], a[2], a[3]), (a[2], 2*a[1], a[4])))
            v = _nullspace(A)[:,0]
            v /= v[2]

            # Only accept fittings where maximum is within half of a pixel of
            # the maximum pixel's centre.
            if np.any(np.abs(v[:2]) > 0.5):
                continue

            x, y = v[:2]
            lm_values.append(a[0]*x*x + a[1]*y*y + a[2]*x*y + a[3]*x + a[4]*y + a[5])
        else:
            x, y = 0, 0
            lm_values.append(X[r,c])

        lm_refined_rows.append(r+y)
        lm_refined_cols.append(c+x)

    return np.array(lm_refined_rows), np.array(lm_refined_cols), np.array(lm_values)



from dtcwt.coeffs import biort, qshift

from pytest import raises

def test_antonini():
    h0o, g0o, h1o, g1o = biort('antonini')
    assert h0o.shape[0] == 9
    assert g0o.shape[0] == 7
    assert h1o.shape[0] == 7
    assert g1o.shape[0] == 9

def test_legall():
    h0o, g0o, h1o, g1o = biort('legall')
    assert h0o.shape[0] == 5
    assert g0o.shape[0] == 3
    assert h1o.shape[0] == 3
    assert g1o.shape[0] == 5

def test_near_sym_a():
    h0o, g0o, h1o, g1o = biort('near_sym_a')
    assert h0o.shape[0] == 5
    assert g0o.shape[0] == 7
    assert h1o.shape[0] == 7
    assert g1o.shape[0] == 5

def test_near_sym_a():
    h0o, g0o, h1o, g1o = biort('near_sym_b')
    assert h0o.shape[0] == 13
    assert g0o.shape[0] == 19
    assert h1o.shape[0] == 19
    assert g1o.shape[0] == 13

def test_qshift_06():
    coeffs = qshift('qshift_06')
    assert len(coeffs) == 8
    for v in coeffs:
        assert v.shape[0] == 10

def test_qshift_a():
    coeffs = qshift('qshift_a')
    assert len(coeffs) == 8
    for v in coeffs:
        assert v.shape[0] == 10

def test_qshift_b():
    coeffs = qshift('qshift_b')
    assert len(coeffs) == 8
    for v in coeffs:
        assert v.shape[0] == 14

def test_qshift_c():
    coeffs = qshift('qshift_c')
    assert len(coeffs) == 8
    for v in coeffs:
        assert v.shape[0] == 16

def test_qshift_d():
    coeffs = qshift('qshift_d')
    assert len(coeffs) == 8
    for v in coeffs:
        assert v.shape[0] == 18

def test_non_exist_biort():
    with raises(IOError):
        biort('this-does-not-exist')

def test_non_exist_qshift():
    with raises(IOError):
        qshift('this-does-not-exist')

def test_wrong_type_a():
    with raises(ValueError):
        biort('qshift_06')

def test_wrong_type_b():
    with raises(ValueError):
        qshift('antonini')

# vim:sw=4:sts=4:et



import itertools
import numpy as np
from six.moves import xrange

from dtcwt.utils import *

def test_complex_type_for_complex():
    assert np.issubsctype(appropriate_complex_type_for(np.zeros((2,3), np.complex64)), np.complex64)
    assert np.issubsctype(appropriate_complex_type_for(np.zeros((2,3), np.complex128)), np.complex128)

def test_complex_type_for_float():
    assert np.issubsctype(appropriate_complex_type_for(np.zeros((2,3), np.float32)), np.complex64)
    assert np.issubsctype(appropriate_complex_type_for(np.zeros((2,3), np.float64)), np.complex128)

def test_draw_circ():
    c = drawcirc(20, 4, -4, -5, 50)
    assert c.shape == (50, 50)
    assert c[45,45] == 0
    assert c[25,25] == 1

def test_draw_edge():
    e = drawedge(20, (20,20), 4, 50)
    assert e.shape == (50, 50)

def test_stacked_2d_matrix_vector_product():
    nr, nc = (30, 20)
    mr, mc = (3, 4)

    # Generate a random stack of 2d matrices and 2d vectors
    matrices = np.random.rand(nr, nc, mr, mc)
    vectors = np.random.rand(nr, nc, mc)

    # Compute product to test
    out = stacked_2d_matrix_vector_prod(matrices, vectors)

    # Check output shape is what we expect
    assert out.shape == (nr, nc, mr)

    # Check each product
    for r, c in itertools.product(xrange(nr), xrange(nc)):
        m = matrices[r, c, :, :]
        v = vectors[r, c, :]
        o = out[r, c, :]

        assert m.shape == (mr, mc)
        assert v.shape == (mc,)
        assert o.shape == (mr,)

        gold = m.dot(v)
        max_delta = np.abs(gold - o).max()
        assert max_delta < 1e-8

def test_stacked_2d_vector_matrix_product():
    nr, nc = (30, 20)
    mr, mc = (3, 4)

    # Generate a random stack of 2d matrices and 2d vectors
    matrices = np.random.rand(nr, nc, mr, mc)
    vectors = np.random.rand(nr, nc, mr)

    # Compute product to test
    out = stacked_2d_vector_matrix_prod(vectors, matrices)

    # Check output shape is what we expect
    assert out.shape == (nr, nc, mc)

    # Check each product
    for r, c in itertools.product(xrange(nr), xrange(nc)):
        m = matrices[r, c, :, :]
        v = vectors[r, c, :]
        o = out[r, c, :]

        assert m.shape == (mr, mc)
        assert v.shape == (mr,)
        assert o.shape == (mc,)

        gold = v.dot(m)
        max_delta = np.abs(gold - o).max()
        assert max_delta < 1e-8

def test_stacked_2d_matrix_matrix_product():
    nr, nc = (30, 20)
    mr1, k, mc2 = (3, 4, 5)

    # Generate a random stack of 2d matrices and 2d vectors
    matrices1 = np.random.rand(nr, nc, mr1, k)
    matrices2 = np.random.rand(nr, nc, k, mc2)

    # Compute product to test
    out = stacked_2d_matrix_matrix_prod(matrices1, matrices2)

    # Check output shape is what we expect
    assert out.shape == (nr, nc, mr1, mc2)

    # Check each product
    for r, c in itertools.product(xrange(nr), xrange(nc)):
        m1 = matrices1[r, c, :, :]
        m2 = matrices2[r, c, :, :]
        o = out[r, c, :, :]

        assert m1.shape == (mr1, k)
        assert m2.shape == (k, mc2)
        assert o.shape == (mr1, mc2)

        gold = m1.dot(m2)
        max_delta = np.abs(gold - o).max()
        assert max_delta < 1e-8




#!/usr/bin/env python
"""Show an example of how to re-sample high-pass DT-CWT coefficients.

"""
import os

import dtcwt
import dtcwt.compat
import dtcwt.sampling

# Use an off-screen backend for matplotlib
import matplotlib
matplotlib.use('agg')

# Import numpy and matplotlib's pyplot interface
import numpy as np
from matplotlib.pyplot import *

# Get a copy of the famous 'mandrill' image. In the default dtcwt tree, we ship
# one with the tests. The mandrill image is 512x512, floating point and has pixel
# values on the interval (0, 1].
mandrill = np.load(
    os.path.join(os.path.dirname(__file__), '..', 'tests', 'mandrill.npz')
)['mandrill']

# Chop a window out
mandrill = mandrill[224:288,224:288]

# We will try to re-scale mandrill by this amount and method
scale = 1.2
scale_method = 'lanczos'

def scale_direct(im):
    """Scale image directly."""
    return dtcwt.sampling.rescale(im, (im.shape[0]*scale, im.shape[1]*scale), scale_method)

def scale_highpass(im):
    """Scale image assuming it to be wavelet highpass coefficients."""
    return dtcwt.sampling.rescale_highpass(im, (im.shape[0]*scale, im.shape[1]*scale), scale_method)

# Rescale mandrill directly using default (Lanczos) sampling
mandrill_direct = scale_direct(mandrill)

# Transform mandrill
mandrill_l, mandrill_h = dtcwt.compat.dtwavexfm2(mandrill, nlevels=4)

# Re-scale each component and transform back. Do this both with and without
# shifting back to DC.
mandrill_l = scale_direct(mandrill_l)
mandrill_h_a, mandrill_h_b = [], []

for h in mandrill_h:
    mandrill_h_a.append(scale_direct(h))
    mandrill_h_b.append(scale_highpass(h))

# Transform back
mandrill_a = dtcwt.compat.dtwaveifm2(mandrill_l, mandrill_h_a)
mandrill_b = dtcwt.compat.dtwaveifm2(mandrill_l, mandrill_h_b)

figure(figsize=(10,10))

subplot(2,2,1)
imshow(mandrill, cmap=cm.gray, clim=(0,1), interpolation='none')
axis('off')
title('Original')

subplot(2,2,2)
imshow(mandrill_direct, cmap=cm.gray, clim=(0,1), interpolation='none')
axis('off')
title('Directly up-sampled')

subplot(2,2,3)
imshow(mandrill_a, cmap=cm.gray, clim=(0,1), interpolation='none')
axis('off')
title('Up-sampled in the wavelet domain')

subplot(2,2,4)
imshow(mandrill_b, cmap=cm.gray, clim=(0,1), interpolation='none')
axis('off')
title('Up-sampled in the wavelet domain with shifting')

tight_layout()
savefig('resampling-example.png')



import os
from pytest import raises

import numpy as np
from dtcwt.compat import dtwavexfm2, dtwaveifm2
from dtcwt.coeffs import biort, qshift
import tests.datasets as datasets

TOLERANCE = 1e-12

def setup():
    global mandrill
    mandrill = datasets.mandrill()

def test_mandrill_loaded():
    assert mandrill.shape == (512, 512)
    assert mandrill.min() >= 0
    assert mandrill.max() <= 1
    assert mandrill.dtype == np.float32

def test_simple():
    Yl, Yh = dtwavexfm2(mandrill)

def test_specific_wavelet():
    Yl, Yh = dtwavexfm2(mandrill, biort=biort('antonini'), qshift=qshift('qshift_06'))

def test_1d():
    Yl, Yh = dtwavexfm2(mandrill[0,:])

def test_3d():
    with raises(ValueError):
        Yl, Yh = dtwavexfm2(np.dstack((mandrill, mandrill)))

def test_simple_w_scale():
    Yl, Yh, Yscale = dtwavexfm2(mandrill, include_scale=True)

    assert len(Yscale) > 0
    for x in Yscale:
        assert x is not None

def test_odd_rows():
    Yl, Yh = dtwavexfm2(mandrill[:509,:])

def test_odd_rows_w_scale():
    Yl, Yh, Yscale = dtwavexfm2(mandrill[:509,:], include_scale=True)

def test_odd_cols():
    Yl, Yh = dtwavexfm2(mandrill[:,:509])

def test_odd_cols_w_scale():
    Yl, Yh, Yscale = dtwavexfm2(mandrill[:509,:509], include_scale=True)

def test_odd_rows_and_cols():
    Yl, Yh = dtwavexfm2(mandrill[:,:509])

def test_odd_rows_and_cols_w_scale():
    Yl, Yh, Yscale = dtwavexfm2(mandrill[:509,:509], include_scale=True)

def test_rot_symm_modified():
    # This test only checks there is no error running these functions, not that they work
    Yl, Yh, Yscale = dtwavexfm2(mandrill, biort='near_sym_b_bp', qshift='qshift_b_bp', include_scale=True)
    Z = dtwaveifm2(Yl, Yh, biort='near_sym_b_bp', qshift='qshift_b_bp')

def test_0_levels():
    Yl, Yh = dtwavexfm2(mandrill, nlevels=0)
    assert np.all(np.abs(Yl - mandrill) < TOLERANCE)
    assert len(Yh) == 0

def test_0_levels_w_scale():
    Yl, Yh, Yscale = dtwavexfm2(mandrill, nlevels=0, include_scale=True)
    assert np.all(np.abs(Yl - mandrill) < TOLERANCE)
    assert len(Yh) == 0
    assert len(Yscale) == 0

def test_integer_input():
    # Check that an integer input is correctly coerced into a floating point
    # array
    Yl, Yh = dtwavexfm2([[1,2,3,4], [1,2,3,4]])
    assert np.any(Yl != 0)

def test_integer_perfect_recon():
    # Check that an integer input is correctly coerced into a floating point
    # array and reconstructed
    A = np.array([[1,2,3,4], [5,6,7,8]], dtype=np.int32)
    Yl, Yh = dtwavexfm2(A)
    B = dtwaveifm2(Yl, Yh)
    assert np.max(np.abs(A-B)) < 1e-5

def test_float32_input():
    # Check that an float32 input is correctly output as float32
    Yl, Yh = dtwavexfm2(mandrill.astype(np.float32))
    assert np.issubsctype(Yl.dtype, np.float32)
    assert np.all(list(np.issubsctype(x.dtype, np.complex64) for x in Yh))

# vim:sw=4:sts=4:et



import os

import numpy as np
from dtcwt.opencl.lowlevel import colifilt
from dtcwt.numpy.lowlevel import colifilt as colifilt_gold
from dtcwt.coeffs import biort, qshift

from pytest import raises

from .util import assert_almost_equal, skip_if_no_cl
import tests.datasets as datasets

def setup():
    global mandrill
    mandrill = datasets.mandrill()

def test_mandrill_loaded():
    assert mandrill.shape == (512, 512)
    assert mandrill.min() >= 0
    assert mandrill.max() <= 1
    assert mandrill.dtype == np.float32

@skip_if_no_cl
def test_odd_filter():
    with raises(ValueError):
        colifilt(mandrill, (-1,2,-1), (-1,2,1))

@skip_if_no_cl
def test_different_size_h():
    with raises(ValueError):
        colifilt(mandrill, (-1,2,1), (-0.5,-1,2,-1,0.5))

@skip_if_no_cl
def test_zero_input():
    Y = colifilt(np.zeros_like(mandrill), (-1,1), (1,-1))
    assert np.all(Y[:0] == 0)

@skip_if_no_cl
def test_bad_input_size():
    with raises(ValueError):
        colifilt(mandrill[:511,:], (-1,1), (1,-1))

@skip_if_no_cl
def test_good_input_size():
    Y = colifilt(mandrill[:,:511], (-1,1), (1,-1))
    Z = colifilt_gold(mandrill[:,:511], (-1,1), (1,-1))
    assert_almost_equal(Y,Z)

@skip_if_no_cl
def test_output_size():
    Y = colifilt(mandrill, (-1,1), (1,-1))
    assert Y.shape == (mandrill.shape[0]*2, mandrill.shape[1])
    Z = colifilt_gold(mandrill, (-1,1), (1,-1))
    assert_almost_equal(Y,Z)

@skip_if_no_cl
def test_non_orthogonal_input():
    Y = colifilt(mandrill, (1,1), (1,1))
    assert Y.shape == (mandrill.shape[0]*2, mandrill.shape[1])
    Z = colifilt_gold(mandrill, (1,1), (1,1))
    assert_almost_equal(Y,Z)

@skip_if_no_cl
def test_output_size_non_mult_4():
    Y = colifilt(mandrill, (-1,0,0,1), (1,0,0,-1))
    assert Y.shape == (mandrill.shape[0]*2, mandrill.shape[1])
    Z = colifilt_gold(mandrill, (-1,0,0,1), (1,0,0,-1))
    assert_almost_equal(Y,Z)

@skip_if_no_cl
def test_non_orthogonal_input_non_mult_4():
    Y = colifilt(mandrill, (1,0,0,1), (1,0,0,1))
    assert Y.shape == (mandrill.shape[0]*2, mandrill.shape[1])
    Z = colifilt_gold(mandrill, (1,0,0,1), (1,0,0,1))
    assert_almost_equal(Y,Z)

@skip_if_no_cl
def test_qshift():
    h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = qshift('qshift_d')
    y = colifilt(mandrill, h1b, h1b)
    z = colifilt_gold(mandrill, h1b, h1a)
    assert_almost_equal(y, z)

# This test fails. I'm not sure if that's expected or not because it is using
# colifilt in an odd way.
#
# @skip_if_no_cl
# def test_qshift_odd_len_input_1():
#     h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = qshift('qshift_d')
#     h1a = h1a[:-2]
#     h1b = h1a[::-1]
#     y = colifilt(mandrill, h1a, h1b)
#     z = colifilt_gold(mandrill, h1a, h1b)
#     assert_almost_equal(y, z)

@skip_if_no_cl
def test_qshift_odd_len_input_2():
    h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = qshift('qshift_d')
    y = colifilt(mandrill, h1a[1:-1], h1b[1:-1])
    z = colifilt_gold(mandrill, h1a[1:-1], h1b[1:-1])
    assert_almost_equal(y, z)

@skip_if_no_cl
def test_qshift_even_input():
    h1b = np.array((-0.25, 0.5, 0.5, -0.25))
    h1a = h1b[::-1]
    y = colifilt(mandrill, h1b, h1a)
    z = colifilt_gold(mandrill, h1b, h1a)
    assert_almost_equal(y, z)

# vim:sw=4:sts=4:et



""" Useful utilities for testing the 2-D DTCWT with synthetic images"""

from __future__ import absolute_import

import functools
import numpy as np


def unpack(pyramid, backend='numpy'):
    """ Unpacks a pyramid give back the constituent parts.

    :param pyramid: The Pyramid of DTCWT transforms you wish to unpack
    :param str backend: A string from 'numpy', 'opencl', or 'tf' indicating
        which attributes you want to unpack from the pyramid.

    :returns: returns a generator which can be unpacked into the Yl, Yh and
        Yscale components of the pyramid. The generator will only return 2
        values if the pyramid was created with the include_scale parameter set
        to false.

    .. note::

        You can still unpack a tf or opencl pyramid as if it were created by a
        numpy. In this case it will return a numpy array, rather than the
        backend specific array type.
    """
    backend = backend.lower()
    if backend == 'numpy':
        yield pyramid.lowpass
        yield pyramid.highpasses
        if pyramid.scales is not None:
            yield pyramid.scales
    elif backend == 'opencl':
        yield pyramid.cl_lowpass
        yield pyramid.cl_highpasses
        if pyramid.cl_scales is not None:
            yield pyramid.cl_scales
    elif backend == 'tf':
        yield pyramid.lowpass_op
        yield pyramid.highpasses_ops
        if pyramid.scales_ops is not None:
            yield pyramid.scales_ops


def drawedge(theta,r,w,N):
    """Generate an image of size N * N pels, of an edge going from 0 to 1
    in height at theta degrees to the horizontal (top of image = 1 if angle = 0).
    r is a two-element vector, it is a coordinate in ij coords through
    which the step should pass.
    The shape of the intensity step is half a raised cosine w pels wide (w>=1).

    T. E . Gale's enhancement to drawedge() for MATLAB, transliterated
    to Python by S. C. Forshaw, Nov. 2013. """

    # convert theta from degrees to radians
    thetar = np.array(theta * np.pi / 180)

    # Calculate image centre from given width
    imCentre = (np.array([N,N]).T - 1) / 2 + 1

    # Calculate values to subtract from the plane
    r = np.array([np.cos(thetar), np.sin(thetar)])*(-1) * (r - imCentre)

    # check width of raised cosine section
    w = np.maximum(1,w)


    ramp = np.arange(0,N) - (N+1)/2
    hgrad = np.sin(thetar)*(-1) * np.ones([N,1])
    vgrad = np.cos(thetar)*(-1) * np.ones([1,N])
    plane = ((hgrad * ramp) - r[0]) + ((ramp * vgrad).T - r[1])
    x = 0.5 + 0.5 * np.sin(np.minimum(np.maximum(plane*(np.pi/w), np.pi/(-2)), np.pi/2))

    return x

def drawcirc(r,w,du,dv,N):

    """Generate an image of size N*N pels, containing a circle
    radius r pels and centred at du,dv relative
    to the centre of the image.  The edge of the circle is a cosine shaped
    edge of width w (from 10 to 90% points).

    Python implementation by S. C. Forshaw, November 2013."""

    # check value of w to avoid dividing by zero
    w = np.maximum(w,1)

    #x plane
    x = np.ones([N,1]) * ((np.arange(0,N,1, dtype='float') - (N+1) / 2 - dv) / r)

    # y vector
    y = (((np.arange(0,N,1, dtype='float') - (N+1) / 2 - du) / r) * np.ones([1,N])).T

    # Final circle image plane
    p = 0.5 + 0.5 * np.sin(np.minimum(np.maximum((np.exp(np.array([-0.5]) * (x**2 + y**2)).T - np.exp((-0.5))) * (r * 3 / w), np.pi/(-2)), np.pi/2))
    return p

def asfarray(X):
    """Similar to :py:func:`numpy.asfarray` except that this function tries to
    preserve the original datatype of X if it is already a floating point type
    and will pass floating point arrays through directly without copying.

    """
    X = np.asanyarray(X)
    return np.asfarray(X, dtype=X.dtype)

def appropriate_complex_type_for(X):
    """Return an appropriate complex data type depending on the type of X. If X
    is already complex, return that, if it is floating point return a complex
    type of the appropriate size and if it is integer, choose an complex
    floating point type depending on the result of :py:func:`numpy.asfarray`.

    """
    X = asfarray(X)

    if np.issubsctype(X.dtype, np.complex64) or np.issubsctype(X.dtype, np.complex128):
        return X.dtype
    elif np.issubsctype(X.dtype, np.float32):
        return np.complex64
    elif np.issubsctype(X.dtype, np.float64):
        return np.complex128

    # God knows, err on the side of caution
    return np.complex128

def as_column_vector(v):
    """Return *v* as a column vector with shape (N,1).

    """
    v = np.atleast_2d(v)
    if v.shape[0] == 1:
        return v.T
    else:
        return v

def reflect(x, minx, maxx):
    """Reflect the values in matrix *x* about the scalar values *minx* and
    *maxx*.  Hence a vector *x* containing a long linearly increasing series is
    converted into a waveform which ramps linearly up and down between *minx* and
    *maxx*.  If *x* contains integers and *minx* and *maxx* are (integers + 0.5), the
    ramps will have repeated max and min samples.

    .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
    .. codeauthor:: Nick Kingsbury, Cambridge University, January 1999.

    """
    x = np.asanyarray(x)
    rng = maxx - minx
    rng_by_2 = 2 * rng
    mod = np.fmod(x - minx, rng_by_2)
    normed_mod = np.where(mod < 0, mod + rng_by_2, mod)
    out = np.where(normed_mod >= rng, rng_by_2 - normed_mod, normed_mod) + minx
    return np.array(out, dtype=x.dtype)

# note that this decorator ignores **kwargs
# From https://wiki.python.org/moin/PythonDecoratorLibrary#Alternate_memoize_as_nested_functions
def memoize(obj):
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        if args not in cache:
            cache[args] = obj(*args, **kwargs)
        return cache[args]
    return memoizer

def stacked_2d_matrix_vector_prod(mats, vecs):
    """
    Interpret *mats* and *vecs* as arrays of 2D matrices and vectors. I.e.
    *mats* has shape PxQxNxM and *vecs* has shape PxQxM. The result
    is a PxQxN array equivalent to:

    .. code::

        result[i,j,:] = mats[i,j,:,:].dot(vecs[i,j,:])

    for all valid row and column indices *i* and *j*.
    """
    return np.einsum('...ij,...j->...i', mats, vecs)

def stacked_2d_vector_matrix_prod(vecs, mats):
    """
    Interpret *mats* and *vecs* as arrays of 2D matrices and vectors. I.e.
    *mats* has shape PxQxNxM and *vecs* has shape PxQxN. The result
    is a PxQxM array equivalent to:

    .. code::

        result[i,j,:] = mats[i,j,:,:].T.dot(vecs[i,j,:])

    for all valid row and column indices *i* and *j*.
    """
    vecshape = np.array(vecs.shape + (1,))
    vecshape[-1:-3:-1] = vecshape[-2:]
    outshape = mats.shape[:-2] + (mats.shape[-1],)
    return stacked_2d_matrix_matrix_prod(vecs.reshape(vecshape), mats).reshape(outshape)

def stacked_2d_matrix_matrix_prod(mats1, mats2):
    """
    Interpret *mats1* and *mats2* as arrays of 2D matrices. I.e.
    *mats1* has shape PxQxNxM and *mats2* has shape PxQxMxR. The result
    is a PxQxNxR array equivalent to:

    .. code::

        result[i,j,:,:] = mats1[i,j,:,:].dot(mats2[i,j,:,:])

    for all valid row and column indices *i* and *j*.
    """
    return np.einsum('...ij,...jk->...ik', mats1, mats2)



import numpy as np
import logging

from six.moves import xrange

from dtcwt.numpy.common import Pyramid
from dtcwt.coeffs import biort as _biort, qshift as _qshift
from dtcwt.defaults import DEFAULT_BIORT, DEFAULT_QSHIFT
from dtcwt.utils import appropriate_complex_type_for, asfarray

from dtcwt.numpy.lowlevel import *

import pdb

class Transform3d(object):
    """
    An implementation of the 3D DT-CWT via NumPy. *biort* and *qshift* are the
    wavelets which parameterise the transform. Valid values are documented in
    :py:func:`dtcwt.coeffs.biort` and :py:func:`dtcwt.coeffs.qshift`.

    """
    def __init__(self, biort=DEFAULT_BIORT, qshift=DEFAULT_QSHIFT, ext_mode=4):
        # Load bi-orthogonal wavelets
        try:
            self.biort = _biort(biort)
        except TypeError:
            self.biort = biort

        # Load quarter sample shift wavelets
        try:
            self.qshift = _qshift(qshift)
        except TypeError:
            self.qshift = qshift

        self.ext_mode = ext_mode

    def forward(self, X, nlevels=3, include_scale=False, discard_level_1=False):
        """Perform a *n*-level DTCWT-3D decompostion on a 3D matrix *X*.

        :param X: 3D real array-like object
        :param nlevels: Number of levels of wavelet decomposition
        :param biort: Level 1 wavelets to use. See :py:func:`dtcwt.coeffs.biort`.
        :param qshift: Level >= 2 wavelets to use. See :py:func:`dtcwt.coeffs.qshift`.
        :param discard_level_1: True if level 1 high-pass bands are to be discarded.

        :returns: a :py:class:`dtcwt.Pyramid` instance

        Each element of the Pyramid *highpasses* tuple is a 4D complex array
        with the 4th dimension having size 28. The 3D slice ``[l][:,:,:,d]``
        corresponds to the complex higpass coefficients for direction d at
        level l where d and l are both 0-indexed.

        If *biort* or *qshift* are strings, they are used as an argument to the
        :py:func:`dtcwt.coeffs.biort` or :py:func:`dtcwt.coeffs.qshift`
        functions. Otherwise, they are interpreted as tuples of vectors giving
        filter coefficients. In the *biort* case, this should be (h0o, g0o,
        h1o, g1o). In the *qshift* case, this should be (h0a, h0b, g0a, g0b,
        h1a, h1b, g1a, g1b).

        There are two values for *ext_mode*, either 4 or 8. If *ext_mode* = 4,
        check whether 1st level is divisible by 2 (if not we raise a
        ``ValueError``). Also check whether from 2nd level onwards, the coefs can
        be divided by 4. If any dimension size is not a multiple of 4, append extra
        coefs by repeating the edges. If *ext_mode* = 8, check whether 1st level is
        divisible by 4 (if not we raise a ``ValueError``). Also check whether from
        2nd level onwards, the coeffs can be divided by 8. If any dimension size is
        not a multiple of 8, append extra coeffs by repeating the edges twice.

        If *discard_level_1* is True the highpass coefficients at level 1 will not be
        discarded. (And, in fact, will never be calculated.) This turns the
        transform from being 8:1 redundant to being 1:1 redundant at the cost of
        no-longer allowing perfect reconstruction. If this option is selected
        then the first element of the *highpasses* tuple will be `None`. Note
        that :py:func:`dtcwt.Transform3d.inverse` will accept the first element
        being `None` and will treat it as being zero.

        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Huizhong Chen, Jan 2009
        .. codeauthor:: Nick Kingsbury, Cambridge University, July 1999.

        """
        X = np.atleast_3d(asfarray(X))

        # If biort has 6 elements instead of 4, then it's a modified
        # rotationally symmetric wavelet
        # FIXME: there's probably a nicer way to do this
        if len(self.biort) == 4:
            h0o, g0o, h1o, g1o = self.biort
        elif len(self.biort) == 6:
            h0o, g0o, h1o, g1o, h2o, g2o = self.biort
        else:
            raise ValueError('Biort wavelet must have 6 or 4 components.')
        # If qshift has 12 elements instead of 8, then it's a modified
        # rotationally symmetric wavelet
        # FIXME: there's probably a nicer way to do this
        if len(self.qshift) == 8:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = self.qshift
        elif len(self.qshift) == 12:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b, h2a, h2b = self.qshift[:10]
        else:
            raise ValueError('Qshift wavelet must have 12 or 8 components.')

        # Check value of ext_mode. TODO: this should really be an enum :S
        if self.ext_mode != 4 and self.ext_mode != 8:
            raise ValueError('ext_mode must be one of 4 or 8')

        Yl = X
        Yh = [None,] * nlevels

        if include_scale:
            # this is only required if the user specifies a third output component.
            Yscale = [None,] * nlevels

        #pdb.set_trace()
        # level is 0-indexed
        for level in xrange(nlevels):
            # Transform
            if level == 0 and discard_level_1:
                Yl = _level1_xfm_no_highpass(Yl, h0o, h1o, self.ext_mode)
            elif level == 0 and not discard_level_1:
                Yl, Yh[level] = _level1_xfm(Yl, h0o, h1o, self.ext_mode)
            else:
                Yl, Yh[level] = _level2_xfm(Yl, h0a, h0b, h1a, h1b, self.ext_mode)
            if include_scale:
                Yscale[level] = Yl.copy()

                #Yh[nlevels+1]=1 #to throw an error for debugging in nose
        if include_scale:
            return Pyramid(Yl, tuple(Yh), tuple(Yscale))
        else:
            return Pyramid(Yl, tuple(Yh))

    def inverse(self, pyramid):
        """Perform an *n*-level dual-tree complex wavelet (DTCWT) 3D
        reconstruction.

        :param pyramid: The :py:class:`dtcwt.Pyramid`-like instance representing the transformed signal.
        :param biort: Level 1 wavelets to use. See :py:func:`biort`.
        :param qshift: Level >= 2 wavelets to use. See :py:func:`qshift`.
        :param ext_mode: Extension mode. See below.

        :returns: Reconstructed real image matrix.

        If *biort* or *qshift* are strings, they are used as an argument to the
        :py:func:`dtcwt.coeffs.biort` or :py:func:`dtcwt.coeffs.qshift`
        functions. Otherwise, they are interpreted as tuples of vectors giving
        filter coefficients. In the *biort* case, this should be (h0o, g0o,
        h1o, g1o). In the *qshift* case, this should be (h0a, h0b, g0a, g0b,
        h1a, h1b, g1a, g1b).

        There are two values for *ext_mode*, either 4 or 8. If *ext_mode* = 4,
        check whether 1st level is divisible by 2 (if not we raise a
        ``ValueError``). Also check whether from 2nd level onwards, the coefs can
        be divided by 4. If any dimension size is not a multiple of 4, append extra
        coefs by repeating the edges. If *ext_mode* = 8, check whether 1st level is
        divisible by 4 (if not we raise a ``ValueError``). Also check whether from
        2nd level onwards, the coeffs can be divided by 8. If any dimension size is
        not a multiple of 8, append extra coeffs by repeating the edges twice.

        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Huizhong Chen, Jan 2009
        .. codeauthor:: Nick Kingsbury, Cambridge University, July 1999.

        """
        Yl = pyramid.lowpass
        Yh = pyramid.highpasses

        # Try to load coefficients if biort is a string parameter
        if len(self.biort) == 4:
            h0o, g0o, h1o, g1o = self.biort
        elif len(self.biort) == 6:
            h0o, g0o, h1o, g1o, h2o, g2o = self.biort
        else:
            raise ValueError('Biort wavelet must have 6 or 4 components.')

        # If qshift has 12 elements instead of 8, then it's a modified
        # rotationally symmetric wavelet
        # FIXME: there's probably a nicer way to do this
        if len(self.qshift) == 8:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = self.qshift
        elif len(self.qshift) == 12:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b, h2a, h2b, g2a, g2b = self.qshift
        else:
            raise ValueError('Qshift wavelet must have 12 or 8 components.')

        X = Yl

        nlevels = len(Yh)
        # level is 0-indexed but interpreted starting from the *last* level
        for level in xrange(nlevels):
            # Transform
            if level == nlevels-1: # non-obviously this is the 'first' level
                if Yh[-level-1] is None:
                    Yl = _level1_ifm_no_highpass(Yl, g0o, g1o)
                else:
                    Yl = _level1_ifm(Yl, Yh[-level-1], g0o, g1o)
            else:
                # Gracefully handle the Yh[0] is None case.
                if Yh[-level-2] is not None:
                    prev_shape = Yh[-level-2].shape
                else:
                    prev_shape = np.array(Yh[-level-1].shape) * 2

                Yl = _level2_ifm(Yl, Yh[-level-1], g0a, g0b, g1a, g1b, self.ext_mode, prev_shape)

        return Yl

def _level1_xfm(X, h0o, h1o, ext_mode):
    """Perform level 1 of the 3d transform.

    """
    # Check shape of input according to ext_mode. Note that shape of X is
    # double original input in each direction.
    if ext_mode == 4 and np.any(np.fmod(X.shape, 2) != 0):
        raise ValueError('Input shape should be a multiple of 2 in each direction when self.ext_mode == 4')
    elif ext_mode == 8 and np.any(np.fmod(X.shape, 4) != 0):
        raise ValueError('Input shape should be a multiple of 4 in each direction when self.ext_mode == 8')

    # Create work area
    work_shape = np.asanyarray(X.shape) * 2

    # We need one extra row per octant if filter length is even
    if h0o.shape[0] % 2 == 0:
        work_shape += 2

    work = np.zeros(work_shape, dtype=X.dtype)

    # Form some useful slices
    s0a = slice(None, work.shape[0] >> 1)
    s1a = slice(None, work.shape[1] >> 1)
    s2a = slice(None, work.shape[2] >> 1)
    s0b = slice(work.shape[0] >> 1, None)
    s1b = slice(work.shape[1] >> 1, None)
    s2b = slice(work.shape[2] >> 1, None)

    x0a = slice(None, X.shape[0])
    x1a = slice(None, X.shape[1])
    x2a = slice(None, X.shape[2])
    x0b = slice(work.shape[0] >> 1, (work.shape[0] >> 1) + X.shape[0])
    x1b = slice(work.shape[1] >> 1, (work.shape[1] >> 1) + X.shape[1])
    x2b = slice(work.shape[2] >> 1, (work.shape[2] >> 1) + X.shape[2])

    # Assign input
    if h0o.shape[0] % 2 == 0:
        work[:X.shape[0], :X.shape[1], :X.shape[2]] = X

        # Copy last rows/cols/slices
        work[ X.shape[0], :X.shape[1], :X.shape[2]] = X[-1, :, :]
        work[:X.shape[0],  X.shape[1], :X.shape[2]] = X[:, -1, :]
        work[:X.shape[0], :X.shape[1],  X.shape[2]] = X[:, :, -1]
        work[X.shape[0], X.shape[1], X.shape[2]] = X[-1,-1,-1]
    else:
        work[s0a, s1a, s2a] = X

    # Loop over 2nd dimension extracting 2D slice from first and 3rd dimensions
    for f in xrange(work.shape[1] >> 1):
        # extract slice
        y = work[s0a, f, x2a].T
        # Do odd top-level filters on 3rd dim. The order here is important
        # since the second filtering will modify the elements of y as well
        # since y is merely a view onto work.
        work[s0a, f, s2b] = colfilter(y, h1o).T
        work[s0a, f, s2a] = colfilter(y, h0o).T

    # Loop over 3rd dimension extracting 2D slice from first and 2nd dimensions
    for f in xrange(work.shape[2]):
        # Do odd top-level filters on rows.
        y1 = work[x0a, x1a, f].T
        y2 = np.vstack((colfilter(y1, h0o), colfilter(y1, h1o))).T

        # Do odd top-level filters on columns.
        work[s0a, :, f] = colfilter(y2, h0o)
        work[s0b, :, f] = colfilter(y2, h1o)
        #if f==2:
        #work[:,:,work.shape[2]+1]=1 #to throw an error so we can inspect y in the unit test

    # Return appropriate slices of output
    return (
        work[s0a, s1a, s2a],                # LLL
        np.concatenate((
            cube2c(work[x0a, x1b, x2a]),    # HLL
            cube2c(work[x0b, x1a, x2a]),    # LHL
            cube2c(work[x0b, x1b, x2a]),    # HHL
            cube2c(work[x0a, x1a, x2b]),    # LLH
            cube2c(work[x0a, x1b, x2b]),    # HLH
            cube2c(work[x0b, x1a, x2b]),    # LHH
            cube2c(work[x0b, x1b, x2b]),    # HHH
            ), axis=3)
        )

def _level1_xfm_no_highpass(X, h0o, h1o, ext_mode):
    """Perform level 1 of the 3d transform discarding highpass subbands.

    """
    # Check shape of input according to ext_mode. Note that shape of X is
    # double original input in each direction.
    if ext_mode == 4 and np.any(np.fmod(X.shape, 2) != 0):
        raise ValueError('Input shape should be a multiple of 2 in each direction when self.ext_mode == 4')
    elif ext_mode == 8 and np.any(np.fmod(X.shape, 4) != 0):
        raise ValueError('Input shape should be a multiple of 4 in each direction when self.ext_mode == 8')

    out = np.zeros_like(X)

    # Loop over 2nd dimension extracting 2D slice from first and 3rd dimensions
    for f in xrange(X.shape[1]):
        # extract slice
        y = X[:, f, :].T
        out[:, f, :] = colfilter(y, h0o).T

  # Loop over 3rd dimension extracting 2D slice from first and 2nd dimensions
    for f in xrange(X.shape[2]):
        y = colfilter(out[:, :, f].T, h0o).T
        out[:, :, f] = colfilter(y, h0o)

    return out

def _level2_xfm(X, h0a, h0b, h1a, h1b, ext_mode):
    """Perform level 2 or greater of the 3d transform.

    """

    if ext_mode == 4:
        if X.shape[0] % 4 != 0:
            X = np.concatenate((X[[0],:,:], X, X[[-1],:,:]), 0)
        if X.shape[1] % 4 != 0:
            X = np.concatenate((X[:,[0],:], X, X[:,[-1],:]), 1)
        if X.shape[2] % 4 != 0:
            X = np.concatenate((X[:,:,[0]], X, X[:,:,[-1]]), 2)
    elif ext_mode == 8:
        if X.shape[0] % 8 != 0:
            X = np.concatenate((X[(0,0),:,:], X, X[(-1,-1),:,:]), 0)
        if X.shape[1] % 8 != 0:
            X = np.concatenate((X[:,(0,0),:], X, X[:,(-1,-1),:]), 1)
        if X.shape[2] % 8 != 0:
            X = np.concatenate((X[:,:,(0,0)], X, X[:,:,(-1,-1)]), 2)

    # Create work area
    work_shape = np.asanyarray(X.shape)
    work = np.zeros(work_shape, dtype=X.dtype)

    # Form some useful slices
    s0a = slice(None, work.shape[0] >> 1)
    s1a = slice(None, work.shape[1] >> 1)
    s2a = slice(None, work.shape[2] >> 1)
    s0b = slice(work.shape[0] >> 1, None)
    s1b = slice(work.shape[1] >> 1, None)
    s2b = slice(work.shape[2] >> 1, None)

    # Assign input
    work = X

    # Loop over 2nd dimension extracting 2D slice from first and 3rd dimensions
    for f in xrange(work.shape[1]):
        # extract slice (copy required because we overwrite the work array)
        y = work[:, f, :].T.copy()

        # Do even Qshift filters on 3rd dim.
        work[:, f, s2b] = coldfilt(y, h1b, h1a).T
        work[:, f, s2a] = coldfilt(y, h0b, h0a).T

    # Loop over 3rd dimension extracting 2D slice from first and 2nd dimensions
    for f in xrange(work.shape[2]):
        # Do even Qshift filters on rows.
        y1 = work[:, :, f].T
        y2 = np.vstack((coldfilt(y1, h0b, h0a), coldfilt(y1, h1b, h1a))).T

        # Do even Qshift filters on columns.
        work[s0a, :, f] = coldfilt(y2, h0b, h0a)
        work[s0b, :, f] = coldfilt(y2, h1b, h1a)

    # Return appropriate slices of output
    return (
        work[s0a, s1a, s2a],                # LLL
        np.concatenate((
            cube2c(work[s0a, s1b, s2a]),    # HLL
            cube2c(work[s0b, s1a, s2a]),    # LHL
            cube2c(work[s0b, s1b, s2a]),    # HHL
            cube2c(work[s0a, s1a, s2b]),    # LLH
            cube2c(work[s0a, s1b, s2b]),    # HLH
            cube2c(work[s0b, s1a, s2b]),    # LHH
            cube2c(work[s0b, s1b, s2b]),    # HHH
            ), axis=3)
        )

def _level1_ifm(Yl, Yh, g0o, g1o):
    """
    Perform level 1 of the inverse 3d transform.
    """
    # Create work area
    work = np.zeros(np.asanyarray(Yl.shape) * 2, dtype=Yl.dtype)

    # Work out shape of output
    Xshape = np.asanyarray(work.shape) >> 1
    if g0o.shape[0] % 2 == 0:
        # if we have an even length filter, we need to shrink the output by 1
        # to compensate for the addition of an extra row/column/slice in
        # the forward transform
        Xshape -= 1

    # Form some useful slices
    s0a = slice(None, work.shape[0] >> 1)
    s1a = slice(None, work.shape[1] >> 1)
    s2a = slice(None, work.shape[2] >> 1)
    s0b = slice(work.shape[0] >> 1, None)
    s1b = slice(work.shape[1] >> 1, None)
    s2b = slice(work.shape[2] >> 1, None)

    x0a = slice(None, Xshape[0])
    x1a = slice(None, Xshape[1])
    x2a = slice(None, Xshape[2])
    x0b = slice(work.shape[0] >> 1, (work.shape[0] >> 1) + Xshape[0])
    x1b = slice(work.shape[1] >> 1, (work.shape[1] >> 1) + Xshape[1])
    x2b = slice(work.shape[2] >> 1, (work.shape[2] >> 1) + Xshape[2])

    # Assign regions of work area
    work[s0a, s1a, s2a] = Yl
    work[x0a, x1b, x2a] = c2cube(Yh[:,:,:, 0:4 ])
    work[x0b, x1a, x2a] = c2cube(Yh[:,:,:, 4:8 ])
    work[x0b, x1b, x2a] = c2cube(Yh[:,:,:, 8:12])
    work[x0a, x1a, x2b] = c2cube(Yh[:,:,:,12:16])
    work[x0a, x1b, x2b] = c2cube(Yh[:,:,:,16:20])
    work[x0b, x1a, x2b] = c2cube(Yh[:,:,:,20:24])
    work[x0b, x1b, x2b] = c2cube(Yh[:,:,:,24:28])

    for f in xrange(work.shape[2]):
        # Do odd top-level filters on rows.
        y = colfilter(work[:, x1a, f].T, g0o) + colfilter(work[:, x1b, f].T, g1o)

        # Do odd top-level filters on columns.
        work[s0a, s1a, f] = colfilter(y[:, x0a].T, g0o) + colfilter(y[:, x0b].T, g1o)

    for f in xrange(work.shape[1]>>1):
        # Do odd top-level filters on 3rd dim.
        y = work[s0a, f, :].T
        work[s0a, f, s2a] = (colfilter(y[x2a, :], g0o) + colfilter(y[x2b, :], g1o)).T

    if g0o.shape[0] % 2 == 0:
        return work[1:(work.shape[0]>>1), 1:(work.shape[1]>>1), 1:(work.shape[2]>>1)]
    else:
        return work[s0a, s1a, s2a]

def _level1_ifm_no_highpass(Yl, g0o, g1o):
    """Perform level 1 of the inverse 3d transform assuming highpass
    coefficients are zero.

    """
    # Create work area
    output = np.zeros_like(Yl)

    for f in xrange(Yl.shape[2]):
        y = colfilter(Yl[:, :, f].T, g0o)
        output[:, :, f] = colfilter(y.T, g0o)

    for f in xrange(Yl.shape[1]):
        y = output[:, f, :].T.copy()
        output[:, f, :] = colfilter(y, g0o)

    return output

def _level2_ifm(Yl, Yh, g0a, g0b, g1a, g1b, ext_mode, prev_level_size):
    """Perform level 2 or greater of the 3d inverse transform.

    """
    # Create work area
    work = np.zeros(np.asanyarray(Yl.shape)*2, dtype=Yl.dtype)

    # Form some useful slices
    s0a = slice(None, work.shape[0] >> 1)
    s1a = slice(None, work.shape[1] >> 1)
    s2a = slice(None, work.shape[2] >> 1)
    s0b = slice(work.shape[0] >> 1, None)
    s1b = slice(work.shape[1] >> 1, None)
    s2b = slice(work.shape[2] >> 1, None)

    # Assign regions of work area
    work[s0a, s1a, s2a] = Yl
    work[s0a, s1b, s2a] = c2cube(Yh[:,:,:, 0:4 ])
    work[s0b, s1a, s2a] = c2cube(Yh[:,:,:, 4:8 ])
    work[s0b, s1b, s2a] = c2cube(Yh[:,:,:, 8:12])
    work[s0a, s1a, s2b] = c2cube(Yh[:,:,:,12:16])
    work[s0a, s1b, s2b] = c2cube(Yh[:,:,:,16:20])
    work[s0b, s1a, s2b] = c2cube(Yh[:,:,:,20:24])
    work[s0b, s1b, s2b] = c2cube(Yh[:,:,:,24:28])

    for f in xrange(work.shape[2]):
        # Do even Qshift filters on rows.
        y = colifilt(work[:, s1a, f].T, g0b, g0a) + colifilt(work[:, s1b, f].T, g1b, g1a)

          # Do even Qshift filters on columns.
        work[:, :, f] = colifilt(y[:, s0a].T, g0b, g0a) + colifilt(y[:,s0b].T, g1b, g1a)

    for f in xrange(work.shape[1]):
        # Do even Qshift filters on 3rd dim.
        y = work[:, f, :].T
        work[:, f, :] = (colifilt(y[s2a, :], g0b, g0a) + colifilt(y[s2b, :], g1b, g1a)).T

    # Now check if the size of the previous level is exactly twice the size of
    # the current level. If YES, this means we have not done the extension in
    # the previous level. If NO, then we have to remove the appended row /
    # column / frame from the previous level DTCWT coefs.

    prev_level_size = np.asarray(prev_level_size)
    curr_level_size = np.asarray(Yh.shape)

    if ext_mode == 4:
        if curr_level_size[0] * 2 != prev_level_size[0]:
            # Discard the top and bottom rows
            work = work[1:-1,:,:]
        if curr_level_size[1] * 2 != prev_level_size[1]:
            # Discard the top and bottom rows
            work = work[:,1:-1,:]
        if curr_level_size[2] * 2 != prev_level_size[2]:
            # Discard the top and bottom rows
            work = work[:,:,1:-1]
    elif ext_mode == 8:
        if curr_level_size[0] * 2 != prev_level_size[0]:
        # Discard the top and bottom rows
            work = work[2:-2,:,:]
        if curr_level_size[1] * 2 != prev_level_size[1]:
        # Discard the top and bottom rows
            work = work[:,2:-2,:]
        if curr_level_size[2] * 2 != prev_level_size[2]:
        # Discard the top and bottom rows
            work = work[:,:,2:-2]

    return work

#==========================================================================================
#                       **********    INTERNAL FUNCTIONS    **********
#==========================================================================================

def cube2c(y):
    """Convert from octets in y to complex numbers in z.

    Arrange pixels from the corners of the quads into
    2 subimages of alternate real and imag pixels.

        e----f
       /|   /|
      a----b |
      | g- | h
      |/   |/
      c----d

    """

    # TODO: check this scaling
    j2 = 0.5 * np.array([1, 1j])

    # This is taken from:
    # Efficient Registration of Nonrigid 3-D Bodies, Huizhong Chen, and Nick Kingsbury, 2012
    # IEEE TRANSACTIONS ON IMAGE PROCESSING, VOL. 21, NO. 1, JANUARY 2012
    # eqs. (6) to (9)

    A = y[0::2, 0::2, 0::2]
    B = y[0::2, 1::2, 0::2]
    C = y[1::2, 0::2, 0::2]
    D = y[1::2, 1::2, 0::2]
    E = y[0::2, 0::2, 1::2]
    F = y[0::2, 1::2, 1::2]
    G = y[1::2, 0::2, 1::2]
    H = y[1::2, 1::2, 1::2]

    # Combine to form subbands
    p = ( A-G-D-F) * j2[0] + ( B-H+C+E) * j2[1]
    q = ( A-G+D+F) * j2[0] + (-B+H+C+E) * j2[1]
    r = ( A+G+D-F) * j2[0] + ( B+H-C+E) * j2[1]
    s = ( A+G-D+F) * j2[0] + (-B-H-C+E) * j2[1]

    #j2[2]=1 #to throw an error
    # Form the 2 subbands in z.
    z = np.concatenate((
        p[:,:,:,np.newaxis],
        q[:,:,:,np.newaxis],
        r[:,:,:,np.newaxis],
        s[:,:,:,np.newaxis],
    ), axis=3)

    return z

def c2cube(z):
    """Convert from complex numbers octets in z to octets in y.

    Undoes cube2c().

        e----f
       /|   /|
      a----b |
      | g- | h
      |/   |/
      c----d

    """

    scale = 0.5

    p = z[:,:,:,0]
    q = z[:,:,:,1]
    r = z[:,:,:,2]
    s = z[:,:,:,3]

    pr, pi = p.real, p.imag #A,E
    qr, qi = q.real, q.imag #B,F
    rr, ri = r.real, r.imag #C,G
    sr, si = s.real, s.imag #D,H

    y = np.zeros(np.asanyarray(z.shape[:3])*2, dtype=z.real.dtype)

    y[0::2, 0::2, 0::2] = ( pr+qr+rr+sr)
    y[1::2, 0::2, 1::2] = (-pr-qr+rr+sr)
    y[1::2, 1::2, 0::2] = (-pr+qr+rr-sr)
    y[0::2, 1::2, 1::2] = (-pr+qr-rr+sr)

    y[0::2, 1::2, 0::2] = ( pi-qi+ri-si)
    y[1::2, 1::2, 1::2] = (-pi+qi+ri-si)
    y[1::2, 0::2, 0::2] = ( pi+qi-ri-si)
    y[0::2, 0::2, 1::2] = ( pi+qi+ri+si)

    return y * scale

# vim:sw=4:sts=4:et



import os
from pytest import raises

import numpy as np
from dtcwt.coeffs import biort, qshift
from dtcwt.compat import dtwavexfm2 as dtwavexfm2_np, dtwaveifm2
from dtcwt.opencl.transform2d import dtwavexfm2 as dtwavexfm2_cl

from .util import assert_almost_equal, skip_if_no_cl
import tests.datasets as datasets

TOLERANCE = 1e-12
GOLD_TOLERANCE = 1e-5

def setup():
    global mandrill
    mandrill = datasets.mandrill()

def test_mandrill_loaded():
    assert mandrill.shape == (512, 512)
    assert mandrill.min() >= 0
    assert mandrill.max() <= 1
    assert mandrill.dtype == np.float32

def _compare_transforms(A, B):
    Yl_A, Yh_A = A
    Yl_B, Yh_B = B
    assert_almost_equal(Yl_A, Yl_B, tolerance=GOLD_TOLERANCE)
    for x, y in zip(Yh_A, Yh_B):
        assert_almost_equal(x, y, tolerance=GOLD_TOLERANCE)

@skip_if_no_cl
def test_simple():
    _compare_transforms(dtwavexfm2_np(mandrill), dtwavexfm2_cl(mandrill))

@skip_if_no_cl
def test_specific_wavelet():
    a = dtwavexfm2_np(mandrill, biort=biort('antonini'), qshift=qshift('qshift_06'))
    b = dtwavexfm2_cl(mandrill, biort=biort('antonini'), qshift=qshift('qshift_06'))
    _compare_transforms(a, b)

@skip_if_no_cl
def test_1d():
    a = dtwavexfm2_np(mandrill[0,:])
    b = dtwavexfm2_cl(mandrill[0,:])
    _compare_transforms(a, b)

@skip_if_no_cl
def test_3d():
    with raises(ValueError):
        Yl, Yh = dtwavexfm2_cl(np.dstack((mandrill, mandrill)))

@skip_if_no_cl
def test_simple_w_scale():
    Yl, Yh, Yscale = dtwavexfm2_cl(mandrill, include_scale=True)

    assert len(Yscale) > 0
    for x in Yscale:
        assert x is not None

@skip_if_no_cl
@skip_if_no_cl
def test_odd_rows():
    a = dtwavexfm2_np(mandrill[:509,:])
    b = dtwavexfm2_cl(mandrill[:509,:])
    _compare_transforms(a, b)

@skip_if_no_cl
def test_odd_cols():
    a = dtwavexfm2_np(mandrill[:,:509])
    b = dtwavexfm2_cl(mandrill[:,:509])
    _compare_transforms(a, b)

@skip_if_no_cl
def test_odd_rows_and_cols():
    a = dtwavexfm2_np(mandrill[:509,:509])
    b = dtwavexfm2_cl(mandrill[:509,:509])
    _compare_transforms(a, b)

@skip_if_no_cl
def test_0_levels():
    a = dtwavexfm2_np(mandrill, nlevels=0)
    b = dtwavexfm2_cl(mandrill, nlevels=0)
    _compare_transforms(a, b)

@skip_if_no_cl
def test_modified():
    a = dtwavexfm2_np(mandrill, biort=biort('near_sym_b_bp'), qshift=qshift('qshift_b_bp'))
    b = dtwavexfm2_cl(mandrill, biort=biort('near_sym_b_bp'), qshift=qshift('qshift_b_bp'))
    _compare_transforms(a, b)

# vim:sw=4:sts=4:et



import os

import numpy as np
from dtcwt.compat import dtwavexfm2, dtwaveifm2
from dtcwt.coeffs import biort, qshift
import tests.datasets as datasets

TOLERANCE = 1e-12

def setup():
    global mandrill, mandrill_crop
    mandrill = datasets.mandrill().astype(np.float64)
    mandrill_crop = mandrill[:233, :301]

def test_mandrill_loaded():
    assert mandrill.shape == (512, 512)
    assert mandrill.min() >= 0
    assert mandrill.max() <= 1
    assert mandrill.dtype == np.float64

def test_reconstruct():
    # Reconstruction up to tolerance
    Yl, Yh = dtwavexfm2(mandrill)
    mandrill_recon = dtwaveifm2(Yl, Yh)
    assert np.all(np.abs(mandrill_recon - mandrill) < TOLERANCE)

def test_reconstruct_crop():
    # Reconstruction up to tolerance
    Yl_crop, Yh_crop = dtwavexfm2(mandrill_crop)
    mandrill_recon = dtwaveifm2(Yl_crop, Yh_crop)[:mandrill_crop.shape[0], :mandrill_crop.shape[1]]
    assert np.all(np.abs(mandrill_recon - mandrill_crop) < TOLERANCE)

def test_reconstruct_custom_filter():
    # Reconstruction up to tolerance
    Yl, Yh = dtwavexfm2(mandrill, 4, biort('legall'), qshift('qshift_06'))
    mandrill_recon = dtwaveifm2(Yl, Yh, biort('legall'), qshift('qshift_06'))
    assert np.all(np.abs(mandrill_recon - mandrill) < TOLERANCE)

def test_float32_input():
    # Check that an float32 input is correctly output as float32
    Yl, Yh = dtwavexfm2(mandrill.astype(np.float32))
    assert np.issubsctype(Yl.dtype, np.float32)
    assert np.all(list(np.issubsctype(x.dtype, np.complex64) for x in Yh))

    mandrill_recon = dtwaveifm2(Yl, Yh)
    assert np.issubsctype(mandrill_recon.dtype, np.float32)


# vim:sw=4:sts=4:et



import os
import sys

__all__ = [
    '__version__',

    'Transform1d',
    'Transform2d',
    'Transform3d',
    'Pyramid',

    'backend_name',
    'push_backend',
    'pop_backend',
    'preserve_backend_stack',
]

from dtcwt._version import __version__

import dtcwt.numpy
import dtcwt.opencl
import dtcwt.tf

# An array of dictionaries. Each dictionary stores the top-level module
# variables for that backend.
_BACKEND_STACK = []

# Possible backends keyed by name
_AVAILABLE_BACKENDS = {
    'numpy': {
        'Transform1d': dtcwt.numpy.Transform1d,
        'Transform2d': dtcwt.numpy.Transform2d,
        'Transform3d': dtcwt.numpy.Transform3d,
        'Pyramid': dtcwt.numpy.Pyramid,
    },
    'opencl': {
        'Transform1d': dtcwt.numpy.Transform1d,
        'Transform2d': dtcwt.opencl.Transform2d,
        'Transform3d': dtcwt.numpy.Transform3d,
        'Pyramid': dtcwt.opencl.Pyramid,
    },
    'tf': {
        'Transform1d': dtcwt.tf.Transform1d,
        'Transform2d': dtcwt.tf.Transform2d,
        'Transform3d': dtcwt.numpy.Transform3d,
        'Pyramid': dtcwt.tf.Pyramid,
    },
}

def _update_from_current_backend():
    # Update values from backend
    for k,v in _BACKEND_STACK[-1][1].items():
        setattr(dtcwt, k, v)
    dtcwt.backend_name = _BACKEND_STACK[-1][0]

class _BackendGuard(object):
    def __init__(self, stack):
        # Explicitly copy the stack
        self._stack = list(stack)

    def __enter__(self):
        return _BACKEND_STACK

    def __exit__(self, exc_type, exc_value, exc_tb):
        dtcwt._BACKEND_STACK = self._stack
        _update_from_current_backend()
        # only re-raise if it's *not* the exception that was
        # passed to throw(), because __exit__() must not raise
        # an exception unless __exit__() itself failed.  But
        # throw() has to raise the exception to signal
        # propagation, so this fixes the impedance mismatch
        # between the throw() protocol and the __exit__()
        # protocol.
        #
        if sys.exc_info()[1] is not exc_value:
            raise

def preserve_backend_stack():
    """Return a generator object which can be used to preserve the backend
    stack even when an exception has been raise. For example:

    .. code-block:: python

        # current backend is NumPy
        assert dtcwt.backend_name == 'numpy'

        with dtcwt.preserve_backend_stack():
            dtcwt.push_backend('opencl')
            # ... things which may raise an exception

        # current backend is NumPy even if an exception was thrown
        assert dtcwt.backend_name == 'numpy'

    """
    return _BackendGuard(_BACKEND_STACK)

def push_backend(name):
    """Switch backend implementation to *name*. Push the previous backend onto
    the backend stack. The previous backend may be restored via
    :py:func:`dtcwt.pop_backend`.

    :param name: string identifying which backend to switch to
    :raises ValueError: if *name* does not correspond to a known backend

    *name* may take one of the following values:

    * ``numpy``: the default NumPy backend. See :py:mod:`dtcwt.numpy`.
    * ``opencl``: a backend which uses OpenCL where available. See
      :py:mod:`dtcwt.opencl`.

    """
    try:
        _BACKEND_STACK.append((name, _AVAILABLE_BACKENDS[name]))
    except KeyError:
        raise ValueError('No such backend: {0}'.format(name))
    _update_from_current_backend()

def pop_backend():
    """Restore the backend after a call to :py:func:`push_backend`. Calls to
    :py:func:`pop_backend` and :py:func:`pop_backend` may be nested. This
    function will undo the most recent call to :py:func:`push_backend`.

    :raise IndexError: if one attempts to pop more backends than one has pushed.

    """
    # It is an error to pop off the default backend
    if len(_BACKEND_STACK) <= 1:
        raise IndexError('Cannot pop default backend')

    _BACKEND_STACK.pop()
    _update_from_current_backend()

backend_name = None
"""A string providing a short human-readable name for the DTCWT backend
currently being used. This corresponds to the *name* parameter passed to
:py:func:`dtcwt.push_backend`. The *default* backend is ``numpy`` but can be
overridden by setting the DTCWT_BACKEND environment variable to a valid backend
name.
"""

# Default to the numpy backend unless DTCWT_BACKEND environment variable is
# set.
push_backend(os.getenv('DTCWT_BACKEND', 'numpy'))



import numpy as np

from dtcwt.utils import reflect

def setup():
    global ramp, reflected

    # Create a simple linear ramp and reflect it
    ramp = np.linspace(-100, 100, 500)
    reflected = reflect(ramp, 30, 40)

def test_linear_ramp_boundaries():
    # Check boundaries
    assert not np.any(reflected < 30)
    assert not np.any(reflected > 40)

def test_linear_ramp_values():
    # Check that valid region is unchanged
    r = np.logical_and(ramp >= 30, ramp <= 40)
    assert np.any(r)
    assert np.all(reflected[r] == ramp[r])

def test_non_array_input():
    ramp = np.linspace(-100, 100, 500).tolist()
    reflected = reflect(ramp, 30, 40)

    # Check boundaries
    assert not np.any(reflected < 30)
    assert not np.any(reflected > 40)


# vim:sw=4:sts=4:et



import os
from pytest import raises

import numpy as np
from dtcwt.compat import dtwavexfm, dtwaveifm

TOLERANCE = 1e-12

def test_reconstruct():
    # Reconstruction up to tolerance
    vec = np.random.rand(630)
    Yl, Yh = dtwavexfm(vec)
    vec_recon = dtwaveifm(Yl, Yh)
    assert np.all(np.abs(vec_recon - vec) < TOLERANCE)

def test_reconstruct_2d():
    # Reconstruction up to tolerance
    vec = np.random.rand(630, 20)
    Yl, Yh = dtwavexfm(vec)
    vec_recon = dtwaveifm(Yl, Yh)
    assert np.all(np.abs(vec_recon - vec) < TOLERANCE)

def test_float32_input():
    # Check that an float32 input is correctly output as float32
    Yl, Yh = dtwavexfm(np.array([1, 2, 3, 4]).astype(np.float32))
    assert np.issubsctype(Yl.dtype, np.float32)
    assert np.all(list(np.issubsctype(x.dtype, np.complex64) for x in Yh))

    recon = dtwaveifm(Yl, Yh)
    assert np.issubsctype(recon.dtype, np.float32)

# vim:sw=4:sts=4:et



from __future__ import absolute_import

import numpy as np
import logging

from six.moves import xrange

from dtcwt.coeffs import biort as _biort, qshift as _qshift
from dtcwt.defaults import DEFAULT_BIORT, DEFAULT_QSHIFT
from dtcwt.numpy.common import Pyramid
from dtcwt.numpy.lowlevel import colfilter, coldfilt, colifilt
from dtcwt.utils import as_column_vector, asfarray

class Transform1d(object):
    """
    An implementation of the 1D DT-CWT in NumPy.

    :param biort: Level 1 wavelets to use. See :py:func:`dtcwt.coeffs.biort`.
    :param qshift: Level >= 2 wavelets to use. See :py:func:`dtcwt.coeffs.qshift`.

    """
    def __init__(self, biort=DEFAULT_BIORT, qshift=DEFAULT_QSHIFT):
        self.biort = biort
        self.qshift = qshift

    def forward(self, X, nlevels=3, include_scale=False):
        """Perform a *n*-level DTCWT decompostion on a 1D column vector *X* (or on
        the columns of a matrix *X*).

        :param X: 1D real array or 2D real array whose columns are to be transformed
        :param nlevels: Number of levels of wavelet decomposition

        :returns: A :py:class:`dtcwt.Pyramid`-like object representing the transform result.

        If *biort* or *qshift* are strings, they are used as an argument to the
        :py:func:`biort` or :py:func:`qshift` functions. Otherwise, they are
        interpreted as tuples of vectors giving filter coefficients. In the *biort*
        case, this should be (h0o, g0o, h1o, g1o). In the *qshift* case, this should
        be (h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b).

        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Nick Kingsbury, Cambridge University, May 2002
        .. codeauthor:: Cian Shaffrey, Cambridge University, May 2002

        """
        # Which wavelets are to be used?
        biort = self.biort
        qshift = self.qshift

        # Need this because colfilter and friends assumes input is 2d
        X = asfarray(X)
        if len(X.shape) == 1:
           X = np.atleast_2d(X).T

        # Try to load coefficients if biort is a string parameter
        try:
            h0o, g0o, h1o, g1o = _biort(biort)
        except TypeError:
            h0o, g0o, h1o, g1o = biort

        # Try to load coefficients if qshift is a string parameter
        try:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = _qshift(qshift)
        except TypeError:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = qshift

        L = np.asanyarray(X.shape)

        # ensure that X is an even length, thus enabling it to be extended if needs be.
        if X.shape[0] % 2 != 0:
            raise ValueError('Size of input X must be a multiple of 2')

        if nlevels == 0:
            if include_scale:
                return Pyramid(X, (), ())
            else:
                return Pyramid(X, ())

        # initialise
        Yh = [None,] * nlevels
        if include_scale:
            # This is only required if the user specifies scales are to be outputted
            Yscale = [None,] * nlevels

        # Level 1.
        Hi = colfilter(X, h1o)
        Lo = colfilter(X, h0o)
        Yh[0] = Hi[::2,:] + 1j*Hi[1::2,:] # Convert Hi to complex form.
        if include_scale:
            Yscale[0] = Lo

        # Levels 2 and above.
        for level in xrange(1, nlevels):
            # Check to see if height of Lo is divisable by 4, if not extend.
            if Lo.shape[0] % 4 != 0:
                Lo = np.vstack((Lo[0,:], Lo, Lo[-1,:]))

            Hi = coldfilt(Lo,h1b,h1a)
            Lo = coldfilt(Lo,h0b,h0a)

            Yh[level] = Hi[::2,:] + 1j*Hi[1::2,:] # Convert Hi to complex form.
            if include_scale:
                Yscale[level] = Lo

        Yl = Lo

        if include_scale:
            return Pyramid(Yl, Yh, Yscale)
        else:
            return Pyramid(Yl, Yh)

    def inverse(self, pyramid, gain_mask=None):
        """Perform an *n*-level dual-tree complex wavelet (DTCWT) 1D
        reconstruction.

        :param pyramid: A :py:class:`dtcwt.Pyramid`-like object containing the transformed signal.
        :param gain_mask: Gain to be applied to each subband.

        :returns: Reconstructed real array.

        The *l*-th element of *gain_mask* is gain for wavelet subband at level l.
        If gain_mask[l] == 0, no computation is performed for band *l*. Default
        *gain_mask* is all ones. Note that *l* is 0-indexed.

        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Nick Kingsbury, Cambridge University, May 2002
        .. codeauthor:: Cian Shaffrey, Cambridge University, May 2002

        """
        # Which wavelets are to be used?
        biort = self.biort
        qshift = self.qshift

        Yl = pyramid.lowpass
        Yh = pyramid.highpasses

        a = len(Yh) # No of levels.

        if gain_mask is None:
            gain_mask = np.ones(a) # Default gain_mask.

        # Try to load coefficients if biort is a string parameter
        try:
            h0o, g0o, h1o, g1o = _biort(biort)
        except TypeError:
            h0o, g0o, h1o, g1o = biort

        # Try to load coefficients if qshift is a string parameter
        try:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = _qshift(qshift)
        except TypeError:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = qshift

        level = a-1   # No of levels = no of rows in L.
        if level < 0:
            # if there are no levels in the input, just return the Yl value
            return Yl

        Lo = Yl
        while level >= 1:  # Reconstruct levels 2 and above in reverse order.
           Hi = c2q1d(Yh[level]*gain_mask[level])
           Lo = colifilt(Lo, g0b, g0a) + colifilt(Hi, g1b, g1a)

           if Lo.shape[0] != 2*Yh[level-1].shape[0]:  # If Lo is not the same length as the next Yh => t1 was extended.
              Lo = Lo[1:-1,...]                       # Therefore we have to clip Lo so it is the same height as the next Yh.

           if np.any(np.asanyarray(Lo.shape) != np.asanyarray(Yh[level-1].shape * np.array((2,1)))):
              raise ValueError('Yh sizes are not valid for DTWAVEIFM')

           level -= 1

        if level == 0:  # Reconstruct level 1.
           Hi = c2q1d(Yh[level]*gain_mask[level])
           Z = colfilter(Lo,g0o) + colfilter(Hi,g1o)

        # Return a 1d vector or a column vector
        if Z.shape[1] == 1:
            return Z.flatten()
        else:
            return Z

#==========================================================================================
#                  **********      INTERNAL FUNCTION    **********
#==========================================================================================

def c2q1d(x):
    """An internal function to convert a 1D Complex vector back to a real
    array,  which is twice the height of x.

    """
    a, b = x.shape
    z = np.zeros((a*2, b), dtype=x.real.dtype)
    z[::2, :] = np.real(x)
    z[1::2, :] = np.imag(x)

    return z

# vim:sw=4:sts=4:et




from __future__ import absolute_import

import numpy as np

from six.moves import xrange

from dtcwt.coeffs import biort as _biort, qshift as _qshift
from dtcwt.defaults import DEFAULT_BIORT, DEFAULT_QSHIFT
from dtcwt.numpy.common import Pyramid as Pyramid_np
from dtcwt.utils import asfarray
from dtcwt.tf import Pyramid
from dtcwt.tf.lowlevel import coldfilt, colfilter, colifilt

try:
    import tensorflow as tf
    from tensorflow.python.framework import dtypes
    tf_dtypes = frozenset(
        [dtypes.float32, dtypes.float64, dtypes.int8, dtypes.int16,
         dtypes.int32, dtypes.int64, dtypes.uint8, dtypes.qint8, dtypes.qint32,
         dtypes.quint8, dtypes.complex64, dtypes.complex128,
         dtypes.float32_ref, dtypes.float64_ref, dtypes.int8_ref,
         dtypes.int16_ref, dtypes.int32_ref, dtypes.int64_ref, dtypes.uint8_ref,
         dtypes.qint8_ref, dtypes.qint32_ref, dtypes.quint8_ref,
         dtypes.complex64_ref, dtypes.complex128_ref]
    )
except ImportError:
    # The lack of tensorflow will be caught by the low-level routines.
    pass

np_dtypes = frozenset(
    [np.dtype('float16'), np.dtype('float32'), np.dtype('float64'),
     np.dtype('int8'), np.dtype('int16'), np.dtype('int32'),
     np.dtype('int64'), np.dtype('uint8'), np.dtype('uint16'),
     np.dtype('uint32'), np.dtype('complex64'), np.dtype('complex128')]
)


class Transform1d(object):
    """
    An implementation of the 1D DT-CWT in Tensorflow.

    :param biort: Level 1 wavelets to use. See :py:func:`dtcwt.coeffs.biort`.
    :param qshift: Level >= 2 wavelets to use. See
        :py:func:`dtcwt.coeffs.qshift`.

    .. note::

        Calling the methods in this class with different inputs will slightly
        vary the results. If you call the
        :py:meth:`~dtcwt.tf.Transform1d.forward` or
        :py:meth:`~dtcwt.tf.Transform1d.forward_channels` methods with a numpy
        array, they load this array into a :py:class:`tf.Variable` and create
        the graph. Subsequent calls to :py:attr:`dtcwt.tf.Pyramid.lowpass` or
        other attributes in the pyramid will create a session and evaluate these
        parameters.  If the above methods are called with a tensorflow variable
        or placeholder, these will be used to create the graph. As such, to
        evaluate the results, you will need to look at the
        :py:attr:`dtcwt.tf.Pyramid.lowpass_op` attribute (calling the `lowpass`
        attribute will try to evaluate the graph with no initialized variables
        and likely result in a runtime error).

        The behaviour is similar for the
        :py:meth:`~dtcwt.tf.Transform1d.inverse` and
        :py:meth:`~dtcwt.tf.Transform1d.inverse_channels` methods, except these
        return an array, rather than a Pyramid style class. If a
        :py:class:`dtcwt.tf.Pyramid` was created by calling the forward methods
        with a numpy array, providing this pyramid to the inverse methods will
        return a numpy array. If however a :py:class:`dtcwt.tf.Pyramid` was
        created by calling the forward methods with a tensorflow variable, the
        result from calling the inverse methods will also be a tensorflow
        variable.
    """
    def __init__(self, biort=DEFAULT_BIORT, qshift=DEFAULT_QSHIFT):
        self.biort = biort
        self.qshift = qshift

    def forward(self, X, nlevels=3, include_scale=False):
        """Perform a *n*-level DTCWT decompostion on a 1D column vector *X* (or on
        the columns of a matrix *X*).

        Can provide the forward transform with either an np array (naive usage),
        or a tensorflow variable or placeholder (designed usage). To transform
        batches of vectors, use the :py:meth:`forward_channels` method.

        :param X: 1D real array or 2D real array whose columns are to be
            transformed.
        :param nlevels: Number of levels of wavelet decomposition

        :returns: A :py:class:`dtcwt.tf.Pyramid` object representing the
            transform result.

        If *biort* or *qshift* are strings, they are used as an argument to the
        :py:func:`biort` or :py:func:`qshift` functions. Otherwise, they are
        interpreted as tuples of vectors giving filter coefficients. In the
        *biort* case, this should be (h0o, g0o, h1o, g1o). In the *qshift* case,
        this should be (h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b).

        .. codeauthor:: Fergal Cotter <fbc23@cam.ac.uk>, Sep 2017
        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Nick Kingsbury, Cambridge University, May 2002
        .. codeauthor:: Cian Shaffrey, Cambridge University, May 2002

        """
        # Check if a numpy array was provided
        numpy = False
        try:
            dtype = X.dtype
        except AttributeError:
            X = asfarray(X)
            dtype = X.dtype

        if dtype in np_dtypes:
            numpy = True
            # Need this because colfilter and friends assumes input is 2d
            if len(X.shape) == 1:
                X = np.atleast_2d(X).T
            X = tf.Variable(X, dtype=tf.float32, trainable=False)
        elif dtype in tf_dtypes:
            if len(X.get_shape().as_list()) == 1:
                X = tf.expand_dims(X, axis=-1)
        else:
            raise ValueError('I cannot handle the variable you have ' +
                             'provided of type ' + str(X.dtype) + '. ' +
                             'Inputs should be a numpy or tf array')

        X_shape = tuple(X.get_shape().as_list())
        size = '{}'.format(X_shape[0])
        name = 'dtcwt_fwd_{}'.format(size)
        if len(X_shape) == 2:
            # Need to make it a batch for tensorflow
            X = tf.expand_dims(X, axis=0)
        elif len(X_shape) >= 3:
            raise ValueError(
                'The entered variable has too many ' +
                'dimensions - ' + str(X_shape) + '.')

        # Do the forward transform
        with tf.variable_scope(name):
            Yl, Yh, Yscale = self._forward_ops(X, nlevels)

        Yl = Yl[0]
        Yh = tuple(x[0] for x in Yh)
        Yscale = tuple(x[0] for x in Yscale)

        if include_scale:
            return Pyramid(Yl, Yh, Yscale, numpy)
        else:
            return Pyramid(Yl, Yh, None, numpy)

    def forward_channels(self, X, nlevels=3, include_scale=False):
        """Perform a *n*-level DTCWT decompostion on a 3D array *X*.

        Can provide the forward transform with either an np array (naive usage),
        or a tensorflow variable or placeholder (designed usage).

        :param X: 3D real array. Batch of matrices whose columns are to be
            transformed (i.e. the second dimension).
        :param nlevels: Number of levels of wavelet decomposition

        :returns: A :py:class:`dtcwt.tf.Pyramid` object representing the
            transform result.

        If *biort* or *qshift* are strings, they are used as an argument to the
        :py:func:`biort` or :py:func:`qshift` functions. Otherwise, they are
        interpreted as tuples of vectors giving filter coefficients. In the
        *biort* case, this should be (h0o, g0o, h1o, g1o). In the *qshift* case,
        this should be (h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b).

        .. codeauthor:: Fergal Cotter <fbc23@cam.ac.uk>, Sep 2017
        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Nick Kingsbury, Cambridge University, May 2002
        .. codeauthor:: Cian Shaffrey, Cambridge University, May 2002

        """
        # Check if a numpy array was provided
        numpy = False
        try:
            dtype = X.dtype
        except AttributeError:
            X = asfarray(X)
            dtype = X.dtype

        if dtype in np_dtypes:
            numpy = True
            if len(X.shape) != 3:
                raise ValueError(
                    'Incorrect input shape for the forward_channels ' +
                    'method ' + str(X.shape) + '. For Inputs of 1 or 2 ' +
                    'dimensions, use the forward method.')
            # Need this because colfilter and friends assumes input is 2d
            X = tf.Variable(X, dtype=tf.float32, trainable=False)
        elif dtype in tf_dtypes:
            X_shape = X.get_shape().as_list()
            if len(X.get_shape().as_list()) != 3:
                raise ValueError(
                    'Incorrect input shape for the forward_channels ' +
                    'method ' + str(X_shape) + '. For Inputs of 1 or 2 ' +
                    'dimensions, use the forward method.')
        else:
            raise ValueError('I cannot handle the variable you have ' +
                             'provided of type ' + str(X.dtype) + '. ' +
                             'Inputs should be a numpy or tf array')

        X_shape = tuple(X.get_shape().as_list())
        size = '{}'.format(X_shape[1])
        name = 'dtcwt_fwd_{}'.format(size)

        # Do the forward transform
        with tf.variable_scope(name):
            Yl, Yh, Yscale = self._forward_ops(X, nlevels)

        if include_scale:
            return Pyramid(Yl, Yh, Yscale, numpy)
        else:
            return Pyramid(Yl, Yh, None, numpy)

    def inverse(self, pyramid, gain_mask=None):
        """Perform an *n*-level dual-tree complex wavelet (DTCWT) 1D
        reconstruction.

        :param pyramid: A :py:class:`dtcwt.Pyramid`-like object containing
            the transformed signal.
        :param gain_mask: Gain to be applied to each subband.

        :returns: Reconstructed real array. Will be a tf Variable if the Pyramid
            was made with tf inputs, otherwise a numpy array.


        The *l*-th element of *gain_mask* is gain for wavelet subband at level
        l.  If gain_mask[l] == 0, no computation is performed for band *l*.
        Default *gain_mask* is all ones. Note that *l* is 0-indexed.

        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Nick Kingsbury, Cambridge University, May 2002
        .. codeauthor:: Cian Shaffrey, Cambridge University, May 2002

        """
        # A tensorflow object was provided
        numpy = False
        if isinstance(pyramid, Pyramid):
            Yl = pyramid.lowpass_op
            Yh = pyramid.highpasses_ops
            numpy = pyramid.numpy

        # Check if a numpy pyramid was provided
        elif isinstance(pyramid, Pyramid_np) or \
                hasattr(pyramid, 'lowpass') and hasattr(pyramid, 'highpasses'):
            numpy = True
            Yl, Yh = pyramid.lowpass, pyramid.highpasses
            Yl = tf.Variable(Yl, trainable=False, dtype=tf.float32)
            Yh = tuple(
                tf.Variable(level, trainable=False, dtype=tf.complex64)
                for level in Yh)
        else:
            raise ValueError(
                'Unknown pyramid provided to inverse transform')

        # Need to make sure it has at least 3 dimensions for tensorflow
        Yl_shape = tuple(Yl.get_shape().as_list())
        if len(Yl_shape) == 2:
            Yl = tf.expand_dims(Yl, axis=0)
            Yh = tuple(tf.expand_dims(x, axis=0) for x in Yh)
        elif len(Yl_shape) >= 3:
            raise ValueError(
                'The entered variables have too many ' +
                'dimensions - ' + str(Yl_shape) + '. For batches of ' +
                'images with multiple channels (i.e. 3 or 4 dimensions), ' +
                'please either enter each channel separately, or use ' +
                'the inverse_channels method.')

        # Do the inverse transform
        s = Yl.get_shape().as_list()[1]
        nlevels = len(Yh)
        size = '{}_up_{}'.format(s, nlevels)
        name = 'dtcwt_inv_{}'.format(size)
        with tf.variable_scope(name):
            X = self._inverse_ops(Yl, Yh, gain_mask)

        # Chop off the first dimension
        X = X[0]

        # Return a 1d vector or a column vector
        if X.get_shape().as_list()[1] == 1:
            X = X[:,0]

        if numpy:
            with tf.Session() as sess:
                sess.run(tf.global_variables_initializer())
                X = sess.run(X)

        return X

    def inverse_channels(self, pyramid, gain_mask=None):
        """Perform an *n*-level dual-tree complex wavelet (DTCWT) 1D
        reconstruction on a 3D array of signals. The inverse is done on the
        second dimension of these.

        This is designed to work after calling the
        :py:meth:`~dtcwt.tf.Transform1d.forward_channels` method.

        :param pyramid: A :py:class:`dtcwt.Pyramid`-like object containing
            the transformed signal. The lowpass signal in the pyramid should be
            a 3D array to use this method.
        :param gain_mask: Gain to be applied to each subband.

        :returns: Reconstructed array. Will be a tf Variable if the Pyramid was
            made with tf inputs, otherwise a numpy array.

        The *l*-th element of *gain_mask* is gain for wavelet subband at level
        l.  If gain_mask[l] == 0, no computation is performed for band *l*.
        Default *gain_mask* is all ones. Note that *l* is 0-indexed.

        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Nick Kingsbury, Cambridge University, May 2002
        .. codeauthor:: Cian Shaffrey, Cambridge University, May 2002

        """
        # A tensorflow object was provided
        numpy = False
        if isinstance(pyramid, Pyramid):
            Yl = pyramid.lowpass_op
            Yl_shape = Yl.get_shape().as_list()
            if len(Yl_shape) != 3:
                raise ValueError(
                    'Incorrect input shape for the forward_channels ' +
                    'method ' + str(Yl_shape) + '. For Inputs of 1 or 2 ' +
                    'dimensions, use the forward method.')
            Yh = pyramid.highpasses_ops
            numpy = pyramid.numpy

        # Check if a numpy pyramid was provided
        elif isinstance(pyramid, Pyramid_np) or \
                hasattr(pyramid, 'lowpass') and hasattr(pyramid, 'highpasses'):
            numpy = True
            Yl, Yh = pyramid.lowpass, pyramid.highpasses
            if len(Yl.shape) != 3:
                raise ValueError(
                    'Incorrect input shape for the forward_channels ' +
                    'method ' + str(Yl.shape) + '. For Inputs of 1 or 2 ' +
                    'dimensions, use the forward method.')

            Yl = tf.Variable(Yl, trainable=False, dtype=tf.float32)
            Yh = tuple(
                tf.Variable(level, trainable=False, dtype=tf.complex64)
                for level in Yh)
        else:
            raise ValueError(
                'Unknown pyramid provided to inverse transform')

        # Do the inverse transform
        s = Yl.get_shape().as_list()[1]
        nlevels = len(Yh)
        size = '{}_up_{}'.format(s, nlevels)
        name = 'dtcwt_inv_{}'.format(size)
        with tf.variable_scope(name):
            X = self._inverse_ops(Yl, Yh, gain_mask)

        if numpy:
            with tf.Session() as sess:
                sess.run(tf.global_variables_initializer())
                X = sess.run(X)

        return X

    def _forward_ops(self, X, nlevels=3):
        """ Perform a *n*-level DTCWT-2D decompostion on a 2D matrix *X*.

        For column inputs, we still need the input shape to be 3D, but with 1 as
        the last dimension.

        :param X: 3D real array of size [batch, h, w]
        :param nlevels: Number of levels of wavelet decomposition
        :param extended: True if a singleton dimension was added at the
            beginning of the input. Signal to remove afterwards.

        :returns: A tuple of Yl, Yh, Yscale
        """
        biort = self.biort
        qshift = self.qshift

        # Try to load coefficients if biort is a string parameter
        try:
            h0o, g0o, h1o, g1o = _biort(biort)
        except TypeError:
            h0o, g0o, h1o, g1o = biort

        # Try to load coefficients if qshift is a string parameter
        try:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = _qshift(qshift)
        except TypeError:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = qshift

        # Check the shape and form of the input
        if X.dtype not in tf_dtypes:
            raise ValueError('X needs to be a tf variable or placeholder')

        original_size = X.get_shape().as_list()[1:]

        # ############################ Resize #################################
        # The next few lines of code check to see if the image is odd in size,
        # if so an extra ... row/column will be added to the bottom/right of the
        # image
        #  initial_row_extend = 0
        #  initial_col_extend = 0
        # If the row count of X is not divisible by 2 then we need to
        # extend X by adding a row at the bottom
        if original_size[0] % 2 != 0:
            #  X = tf.pad(X, [[0, 0], [0, 1], [0, 0]], 'SYMMETRIC')
            raise ValueError('Size of input X must be a multiple of 2')

        #  extended_size = X.get_shape().as_list()[1:]

        if nlevels == 0:
            return X, (), ()

        # ########################### Initialise ###############################
        Yh = [None, ] * nlevels
        # This is only required if the user specifies a third output
        # component.
        Yscale = [None, ] * nlevels

        # ############################ Level 1 #################################
        # Uses the biorthogonal filters
        if nlevels >= 1:
            # Do odd top-level filters on cols.
            Hi = colfilter(X, h1o)
            Lo = colfilter(X, h0o)

            # Convert Hi to complex form by taking alternate rows
            Yh[0] = tf.cast(Hi[:,::2,:], tf.complex64) + \
                1j*tf.cast(Hi[:,1::2,:], tf.complex64)
            Yscale[0] = Lo

        # ############################ Level 2+ ################################
        # Uses the qshift filters
        for level in xrange(1, nlevels):
            # If the row count of Lo is not divisible by 4 (it will be
            # divisible by 2), add 2 extra rows to make it so
            if Lo.get_shape().as_list()[1] % 4 != 0:
                Lo = tf.pad(Lo, [[0, 0], [1, 1], [0, 0]], 'SYMMETRIC')

            # Do even Qshift filters on cols.
            Hi = coldfilt(Lo, h1b, h1a)
            Lo = coldfilt(Lo, h0b, h0a)

            # Convert Hi to complex form by taking alternate rows
            Yh[level] = tf.cast(Hi[:,::2,:], tf.complex64) + \
                1j * tf.cast(Hi[:,1::2,:], tf.complex64)
            Yscale[level] = Lo

        Yl = Lo

        return Yl, tuple(Yh), tuple(Yscale)

    def _inverse_ops(self, Yl, Yh, gain_mask=None):
        """Perform an *n*-level dual-tree complex wavelet (DTCWT) 1D
        reconstruction.

        :param Yl: The lowpass output from a forward transform. Should be a
            tensorflow variable.
        :param Yh: The tuple of highpass outputs from a forward transform.
            Should be tensorflow variables.
        :param gain_mask: Gain to be applied to each subband.

        :returns: A tf.Variable holding the output

        The *l*-th element of *gain_mask* is gain for wavelet subband at level
        l.  If gain_mask[l] == 0, no computation is performed for band *l*.
        Default *gain_mask* is all ones. Note that *l* is 0-indexed.

        .. codeauthor:: Fergal Cotter <fbc23@cam.ac.uk>, Sep 2017
        .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
        .. codeauthor:: Nick Kingsbury, Cambridge University, May 2002
        .. codeauthor:: Cian Shaffrey, Cambridge University, May 2002

        """
        # Which wavelets are to be used?
        biort = self.biort
        qshift = self.qshift
        a = len(Yh)  # No of levels.

        if gain_mask is None:
            gain_mask = np.ones(a)  # Default gain_mask.
        gain_mask = np.array(gain_mask)

        # Try to load coefficients if biort is a string parameter
        try:
            h0o, g0o, h1o, g1o = _biort(biort)
        except TypeError:
            h0o, g0o, h1o, g1o = biort

        # Try to load coefficients if qshift is a string parameter
        try:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = _qshift(qshift)
        except TypeError:
            h0a, h0b, g0a, g0b, h1a, h1b, g1a, g1b = qshift

        level = a-1   # No of levels = no of rows in L.
        if level < 0:
            # if there are no levels in the input, just return the Yl value
            return Yl

        # Reconstruct levels 2 and above in reverse order.
        Lo = Yl
        while level >= 1:
            Hi = c2q1d(Yh[level]*gain_mask[level])
            Lo = colifilt(Lo, g0b, g0a) + colifilt(Hi, g1b, g1a)

            # If Lo is not the same length as the next Therefore we have to clip
            # Lo so it is the same height as the next Yh. Yh => t1 was extended.
            Lo_shape = Lo.get_shape().as_list()
            next_shape = Yh[level-1].get_shape().as_list()
            if Lo_shape[1] != 2 * next_shape[1]:
                Lo = Lo[:,1:-1]
                Lo_shape = Lo.get_shape().as_list()

            # Check the row shapes across the entire matrix
            if (np.any(np.asanyarray(Lo_shape[1:]) !=
                       np.asanyarray(next_shape[1:] * np.array((2,1))))):
                raise ValueError('Yh sizes are not valid for DTWAVEIFM')

            level -= 1

        # Reconstruct level 1.
        if level == 0:
            Hi = c2q1d(Yh[level]*gain_mask[level])
            Z = colfilter(Lo,g0o) + colfilter(Hi,g1o)

        return Z


# =============================================================================
#              **********      INTERNAL FUNCTION    **********
# =============================================================================
def c2q1d(x):
    """ An internal function to convert a 1D Complex vector back to a real
    array,  which is twice the height of x.
    """
    # Input has shape [batch, r, c, 2]
    r, c = x.get_shape().as_list()[1:3]
    x1 = tf.real(x)
    x2 = tf.imag(x)
    # Stack 2 inputs of shape [batch, r, c] to [batch, r, 2, c]
    y = tf.stack([x1, x2], axis=-2)
    # Reshaping interleaves the results
    y = tf.reshape(y, [-1, 2 * r, c])

    return y

# vim:sw=4:sts=4:et



"""
A backend which uses NumPy to perform the filtering. This backend should always
be available.

"""

from .common import Pyramid
from .transform1d import Transform1d
from .transform2d import Transform2d
from .transform3d import Transform3d

__all__ = [
    'Pyramid',
    'Transform1d',
    'Transform2d',
    'Transform3d',
]



"""
.. note::
  This module is experimental. It's API may change between versions.

This module implements function for DTCWT-based image registration as outlined in
`[1] <http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=5936113>`_.
These functions are 2D-only for the moment.

"""
from __future__ import division, absolute_import

import itertools

from six.moves import xrange

import dtcwt
import dtcwt.numpy
import dtcwt.sampling
import dtcwt.utils
import numpy as np

__all__ = [
    'estimatereg',
    'velocityfield',
    'warp',
    'warptransform',
]

# Horizontal and vertical expected phase shifts for each subband
EXPECTED_SHIFTS = np.array(((-1,-3),(-3,-3),(-3,-1),(-3,1),(-3,3),(-1,3))) * np.pi / 2.15

def phasegradient(sb1, sb2, w=None):
    """
    Compute the dx, dy and dt phase gradients for subbands *sb1* and *sb2*.
    Each subband should be an NxM matrix of complex values and should be the
    same size.

    *w* should be a pair giving the expected phase shift for horizontal and vertical
    directions. This expected phase shift is used to 'de-rotate' the phase gradients
    before computing the angle and is added afterwards. If not specified, 0 is
    used. This is likely to give poor results.

    Returns 3 NxM matrices giving the d/dy, d/dx and d/dt phase values
    respectively at position (x,y).

    """
    if w is None:
        w = (0,0)

    if sb1.size != sb2.size:
        raise ValueError('Subbands should have identical size')

    # Measure horizontal phase gradients by taking the angle of
    # summed conjugate products across horizontal pairs.
    S = (sb1[:,1:] * np.conj(sb1[:,:-1]) + sb2[:,1:] * np.conj(sb2[:,:-1])) * np.exp(-1j * w[0])
    #dx = np.angle(dtcwt.sampling.sample(S, xs-0.5, ys, method='bilinear')) + w[0]
    dx = np.hstack((
        np.angle(S[:,:1]),
        np.angle(0.5 * (S[:,:-1] + S[:,1:])),
        np.angle(S[:,-1:])
    )) + w[0]

    # Measure vertical phase gradients by taking the angle of
    # summed conjugate products across vertical pairs.
    S = (sb1[1:,:] * np.conj(sb1[:-1,:]) + sb2[1:,:] * np.conj(sb2[:-1,:])) * np.exp(-1j * w[1])
    dy = np.vstack((
        np.angle(S[:1,:]),
        np.angle(0.5 * (S[:-1,:] + S[1:,:])),
        np.angle(S[-1:,:])
    )) + w[1]

    # Measure temporal phase differences between refh and prevh
    A = sb2 * np.conj(sb1)
    dt = np.angle(A)

    return dy, dx, dt

def _pow2(a):
    return a * a

def _pow3(a):
    return a * a * a

def confidence(sb1, sb2, epsilon=1e-6):
    """
    Compute the confidence measure of subbands *sb1* and *sb2* which should be
    2d arrays of identical shape. The confidence measure gives highest weight
    to pixels we expect to give good results.

    epsilon is a small constant intended to regularise the situation in which
    the case when the wavelet coefficients are very small (and so dominated
    by noise). It should be set to be comparable to the cube of the amplitude
    of the measurement noise.
    """

    if sb1.size != sb2.size:
        raise ValueError('Subbands should have identical size')

    numerator, denominator = 0.0, epsilon

    # Pad subbands
    us = np.concatenate((
        np.concatenate((sb1[  :1,:1], sb1[  :1,:], sb1[  :1,-1:]), axis=1),
        np.concatenate((sb1[  : ,:1], sb1        , sb1[  : ,-1:]), axis=1),
        np.concatenate((sb1[-1: ,:1], sb1[-1: ,:], sb1[-1: ,-1:]), axis=1),
    ), axis=0)
    vs = np.concatenate((
        np.concatenate((sb2[  :1,:1], sb2[  :1,:], sb2[  :1,-1:]), axis=1),
        np.concatenate((sb2[  : ,:1], sb2        , sb2[  : ,-1:]), axis=1),
        np.concatenate((sb2[-1: ,:1], sb2[-1: ,:], sb2[-1: ,-1:]), axis=1),
    ), axis=0)

    us3_abs, vs3_abs = _pow3(np.abs(us)), _pow3(np.abs(vs))
    prod_coeffs = np.conj(us) * vs

    # pixels at -1, -1
    region = (slice(0,-2), slice(0,-2))
    numerator += prod_coeffs[region]
    denominator += us3_abs[region] + vs3_abs[region]

    # pixels at +1, -1
    region = (slice(0,-2), slice(2,None))
    numerator += prod_coeffs[region]
    denominator += us3_abs[region] + vs3_abs[region]

    # pixels at -1, +1
    region = (slice(2,None), slice(0,-2))
    numerator += prod_coeffs[region]
    denominator += us3_abs[region] + vs3_abs[region]

    # pixels at +1, +1
    region = (slice(2,None), slice(2,None))
    numerator += prod_coeffs[region]
    denominator += us3_abs[region] + vs3_abs[region]

    return _pow2(np.abs(numerator)) / denominator

Q_TRIU_INDICES = list(zip(*np.triu_indices(6)))
Q_TRIU_FLAT_INDICES = np.ravel_multi_index(np.triu_indices(6), (6,6))

def qtildematrices(t_ref, t_target, levels):
    r"""
    Compute :math:`\tilde{Q}` matrices for given levels.

    :param t_ref: the transformed reference image
    :param t_target: the transformed target image
    :param levels: a sequence of indices specifying which levels to examine
    :returns: a tuple of :math:`\tilde{Q}` matrices for each index in *levels*

    Both *t_ref* and *t_target* should be
    :py:class:`dtcwt.Pyramid`-compatible objects.
    Indices in *levels* are 0-based.

    The returned matrices are NxMx27 where NxM is the shape of the
    corresponding level's highpass subbands.

    """
    # Extract highpasses
    Yh1 = t_ref.highpasses
    Yh2 = t_target.highpasses

    # A list of arrays of \tilde{Q} matrices for each level
    Qt_mats = []

    for level in levels:
        highpasses1, highpasses2 = Yh1[level], Yh2[level]
        xs, ys = np.meshgrid(np.arange(0,1,1/highpasses1.shape[1]),
                             np.arange(0,1,1/highpasses1.shape[0]))

        Qt_mat_sum = None
        for subband in xrange(highpasses1.shape[2]):
            C_d = confidence(highpasses1[:,:,subband], highpasses2[:,:,subband])
            dy, dx, dt = phasegradient(highpasses1[:,:,subband], highpasses2[:,:,subband],
                                            EXPECTED_SHIFTS[subband,:])

            dx *= highpasses1.shape[1]
            dy *= highpasses1.shape[0]

            # This is the equivalent of the following for each member of the array
            #  Kt_mat = np.array(((1, 0, s*x, 0, s*y, 0, 0), (0, 1, 0, s*x, 0, s*y, 0), (0,0,0,0,0,0,1)))
            #  c_vec = np.array((dx, dy, -dt))
            #  tmp = (Kt_mat.T).dot(c_vec)
            tmp = (
                dx, dy, xs*dx, xs*dy, ys*dx, ys*dy, -dt
            )

            # Calculate Qmatrix elements
            Qt = np.zeros(dx.shape[:2] + (27,))
            elem_idx = 0

            # Q sub-matrix
            for r, c in Q_TRIU_INDICES:
                Qt[:,:,elem_idx] = tmp[r] * tmp[c]
                elem_idx += 1

            # q sub-vector
            for r in xrange(6):
                Qt[:,:,elem_idx] = tmp[r] * tmp[6]
                elem_idx += 1

            # Include the confidence parameter
            Qt *= C_d.reshape(Qt.shape[:-1] + (1,))**2

            # Update Qt mats
            if Qt_mat_sum is None:
                Qt_mat_sum = Qt
            else:
                Qt_mat_sum += Qt

        Qt_mats.append(Qt_mat_sum)

    return Qt_mats

def solvetransform(Qtilde_vec):
    r"""
    Solve for affine transform parameter vector :math:`a` from :math:`\mathbf{\tilde{Q}}`
    matrix. decomposes :math:`\mathbf{\tilde{Q}}` as

    .. math::
        \tilde{Q} = \begin{bmatrix}
            \mathbf{Q}   & \mathbf{q}   \\
            \mathbf{q}^T & q_0
        \end{bmatrix}

    Returns :math:`\mathbf{a} = -\mathbf{Q}^{-1} \mathbf{q}`.
    """

    # Convert from 27-element vector into Q matrix and vector
    Q = np.zeros(Qtilde_vec.shape[:-1] + (6*6,))
    Q[..., Q_TRIU_FLAT_INDICES] = Qtilde_vec[...,:21]
    q = Qtilde_vec[...,-6:]
    Q = np.reshape(Q, Qtilde_vec.shape[:-1] + (6,6))

    # Want to find a = -Q^{-1} q => Qa = -q
    # The naive way would be: return -np.linalg.inv(Q).dot(q)
    # A less naive way would be: return np.linalg.solve(Q, -q)

    # NumPy >1.8 directly supports using the last two dimensions as a matrix. IF
    # we get a LinAlgError, we assume that we need to fall-back to a NumPy 1.7
    # approach which is *significantly* slower.
    try:
        rv = np.linalg.solve(Q, -q)
    except np.linalg.LinAlgError:
        # Try the slower fallback
        rv = np.zeros(Qtilde_vec.shape[:-1] + (6,))
        for idx in itertools.product(*list(xrange(s) for s in Qtilde_vec.shape[:-1])):
            rv[idx] = np.linalg.solve(Q[idx], -q[idx])

    return rv

def normsamplehighpass(Yh, xs, ys, method=None):
    """
    Given a NxMx6 array of subband responses, sample from co-ordinates *xs* and
    *ys*.

    .. note::
      The sample co-ordinates are in *normalised* co-ordinates such that the
      image width and height are both unity.

    """
    return dtcwt.sampling.sample_highpass(Yh, xs*Yh.shape[1], ys*Yh.shape[0], method=method)

def normsample(Yh, xs, ys, method=None):
    """
    Given a NxM image sample from co-ordinates *xs* and *ys*.

    .. note::
      The sample co-ordinates are in *normalised* co-ordinates such that the
      image width and height are both unity.

    """
    return dtcwt.sampling.sample(Yh, xs*Yh.shape[1], ys*Yh.shape[0], method=method)

def warptransform(t, avecs, levels, method=None):
    """
    Return a warped version of a transformed image acting only on specified levels.

    :param t: a transformed image
    :param avecs: an array of affine distortion parameters
    :param levels: a sequence of 0-based indices specifying which levels to act on

    *t* should be a
    :py:class:`dtcwt.Pyramid`-compatible instance.

    The *method* parameter is interpreted as in :py:func:`dtcwt.sampling.rescale` and
    is the sampling method used to resize *avecs* to *shape*.

    .. note::

        This function will clone the transform *t* but it is a shallow clone
        where possible. Only the levels specified in *levels* will be
        deep-copied and warped.

    """
    warped_highpasses = list(t.highpasses)

    # Warp specified levels
    for l in levels:
        warped_highpasses[l] = warphighpass(warped_highpasses[l], avecs, method=method)

    # Clone the transform
    return dtcwt.numpy.Pyramid(t.lowpass, tuple(warped_highpasses), t.scales)

def estimatereg(source, reference, regshape=None, levels=None):
    """
    Estimate registration from which will map *source* to *reference*.

    :param source: transformed source image
    :param reference: transformed reference image

    The *reference* and *source* parameters should support the same API as
    :py:class:`dtcwt.Pyramid`.

    The local affine distortion is estimated at at 8x8 pixel scales.
    Return a NxMx6 array where the 6-element vector at (N,M) corresponds to the
    affine distortion parameters for the 8x8 block with index (N,M).

    Use the :py:func:`velocityfield` function to convert the return value from
    this function into a velocity field.

    If not-`None`, *levels* is a sequence of sequences of 0-based level indices
    to use when calculating the registration. If `None` then a default set of
    levels are used.

    """
    # Extract number of levels and shape of level 4 (i.e. index 3) subband
    nlevels = len(source.highpasses)
    if regshape is None:
        avecs_shape = source.highpasses[3].shape[:2] + (6,)
    else:
        avecs_shape = tuple(regshape[:2]) + (6,)

    # Initialise matrix of 'a' vectors
    avecs = np.zeros(avecs_shape)

    if levels is None:
        levels = []
        levels.append(list(x for x in xrange(nlevels-1, nlevels-3, -1) if x>=0))
        for s in np.arange(nlevels-1, 0, -0.5):
            refine_levels = list(int(np.floor(s))-x for x in range(2) if s-x >= 2)
            if len(refine_levels) < 2:
                continue
            levels.append(refine_levels)

    # Compute initial global transform
    Qt_mats = list(
            np.sum(np.sum(x, axis=0), axis=0)
            for x in qtildematrices(source, reference, levels[0])
    )
    Qt = np.sum(Qt_mats, axis=0)

    a = solvetransform(Qt)
    for idx in xrange(a.shape[0]):
        avecs[:,:,idx] = a[idx]

    # Refine estimate
    for est_levels in levels[1:]:
        # Warp the levels we'll be looking at with the current best-guess transform
        warped = warptransform(source, avecs, est_levels, method='bilinear')

        # Rescale and sample all the Qtilde matrix results
        all_qts = qtildematrices(warped, reference, est_levels)
        if all_qts is None or len(all_qts) < 1:
            continue

        qts = np.zeros(avecs.shape[:2] + all_qts[0].shape[2:])
        for x in all_qts:
            qts += dtcwt.sampling.rescale(_boxfilter(x, 3), avecs.shape[:2], method='bilinear')

        avecs += solvetransform(qts)

    return avecs

def velocityfield(avecs, shape, method=None):
    """
    Given the affine distortion parameters returned from :py:func:`estimatereg`, return
    a tuple of 2D arrays giving the x- and y- components of the velocity field. The
    shape of the velocity component field is *shape*. The velocities are measured in
    terms of normalised units where the image has width and height of unity.

    The *method* parameter is interpreted as in :py:func:`dtcwt.sampling.rescale` and
    is the sampling method used to resize *avecs* to *shape*.

    """
    h, w = avecs.shape[:2]
    pxs, pys = np.meshgrid(np.arange(0, w, dtype=np.float32) / w,
                           np.arange(0, h, dtype=np.float32) / h)

    vxs = avecs[:,:,0] + avecs[:,:,2] * pxs + avecs[:,:,4] * pys
    vys = avecs[:,:,1] + avecs[:,:,3] * pxs + avecs[:,:,5] * pys

    vxs = dtcwt.sampling.rescale(vxs, shape, method=method)
    vys = dtcwt.sampling.rescale(vys, shape, method=method)

    return vxs, vys

def warphighpass(Yh, avecs, method=None):
    """
    A convenience function to warp a highpass subband image according to the
    velocity field implied by *avecs*.

    This function correctly 'de-rotates' the highpass image before sampling and
    're-rotates' afterwards.

    """
    X, Y = np.meshgrid(np.arange(Yh.shape[1], dtype=np.float32)/Yh.shape[1],
                       np.arange(Yh.shape[0], dtype=np.float32)/Yh.shape[0])
    vxs, vys = velocityfield(avecs, Yh.shape, method=method)
    return normsamplehighpass(Yh, X+vxs, Y+vys, method=method)

def warp(I, avecs, method=None):
    """
    A convenience function to warp an image according to the velocity field
    implied by *avecs*.

    """
    X, Y = np.meshgrid(np.arange(I.shape[1], dtype=np.float32)/I.shape[1],
                       np.arange(I.shape[0], dtype=np.float32)/I.shape[0])
    vxs, vys = velocityfield(avecs, I.shape, method=method)
    return normsample(I, X+vxs, Y+vys, method=method)

def _boxfilter(X, kernel_size):
    """
    INTERNAL

    A simple box filter implementation.

    """
    if kernel_size % 2 == 0:
        raise ValueError('Kernel size must be odd')

    for axis_idx in xrange(2):
        slices = [slice(None),] * len(X.shape)
        out = X

        for delta in xrange(1, 1+(kernel_size-1)//2):
            slices[axis_idx] = dtcwt.utils.reflect(
                    np.arange(X.shape[axis_idx]) + delta, -0.5, X.shape[axis_idx]-0.5)
            out = out + X[slices]
            slices[axis_idx] = dtcwt.utils.reflect(
                    np.arange(X.shape[axis_idx]) - delta, -0.5, X.shape[axis_idx]-0.5)
            out = out + X[slices]

        X = out / kernel_size

    return X



#!/usr/bin/env python

import numpy as np
from scipy.io import loadmat
from numpy import savez_compressed

def _mean(a, axis=None, *args, **kwargs):
    """Equivalent to numpy.mean except that the axis along which the mean is taken is not removed."""

    rv = np.mean(a, axis=axis, *args, **kwargs)

    if axis is not None:
        rv = np.expand_dims(rv, axis)

    return rv

def centre_indices(ndim=2,apron=8):
    """Returns the centre indices for the correct number of dimension
    """
    return tuple([slice(apron,-apron) for i in range(ndim)])

def summarise_mat(M, apron=8):
    """HACK to provide a 'summary' matrix consisting of the corners of the
    matrix and summed versions of the sub matrices.

    N.B. Keep this in sync with matlab/verif_m_to_npz.py.

    """
    centre = M[apron:-apron,apron:-apron,...]
    centre_sum = _mean(_mean(centre, axis=0), axis=1)

    return np.vstack((
        np.hstack((M[:apron,:apron,...], _mean(M[:apron,apron:-apron,...], axis=1), M[:apron,-apron:,...])),
        np.hstack((_mean(M[apron:-apron,:apron,...], axis=0), centre_sum, _mean(M[apron:-apron,-apron:,...], axis=0))),
        np.hstack((M[-apron:,:apron,...], _mean(M[-apron:,apron:-apron,...], axis=1), M[-apron:,-apron:,...])),
    ))

def summarise_cube(M, apron=4):
    """Provide a summary cube, extending  summarise_mat to 3D
    """
    return np.dstack(
        [summarise_mat(M[:,:,i,...], apron) for i in range(M.shape[-2])]
    )

verif_temp = loadmat('verification.mat')
verif = dict((k,v) for k, v in verif_temp.items() if (not k.startswith('_') and not k.startswith('qbgn')))
verif_cube = dict((k,v) for k, v in verif_temp.items() if (not k.startswith('_') and k.startswith('qbgn')))
del verif_temp

for idx, v in enumerate(verif['mandrill_Yh']):
    verif['mandrill_Yh_{0}'.format(idx)] = v[0]
del verif['mandrill_Yh']

for idx, v in enumerate(verif['mandrill_Yscale']):
    verif['mandrill_Yscale_{0}'.format(idx)] = v[0]
del verif['mandrill_Yscale']

for idx, v in enumerate(verif['mandrill_Yhb']):
    verif['mandrill_Yhb_{0}'.format(idx)] = v[0]
del verif['mandrill_Yhb']

for idx, v in enumerate(verif['mandrill_Yscaleb']):
    verif['mandrill_Yscaleb_{0}'.format(idx)] = v[0]
del verif['mandrill_Yscaleb']

for idx, v in enumerate(verif_cube['qbgn_Yh']):
    verif_cube['qbgn_Yh_{0}'.format(idx)] = v[0]
del verif_cube['qbgn_Yh']

for idx, v in enumerate(verif_cube['qbgn_Yscale']):
    verif_cube['qbgn_Yscale_{0}'.format(idx)] = v[0]
del verif_cube['qbgn_Yscale']

summaries = dict((k, summarise_mat(v)) for k, v in verif.items())
for k,v in verif_cube.items():
    summaries[k] = summarise_cube(v)

savez_compressed('../tests/verification.npz', **summaries)

# Convert qbgn.mat -> qbgn.npz
savez_compressed('../tests/qbgn.npz', **loadmat('qbgn.mat'))
